from __future__ import annotations

from collections.abc import Iterable
from dataclasses import KW_ONLY, dataclass
from typing import Literal

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["Grid"]


@dataclass
class GridChildPosition:
    row: int
    column: int
    width: int = 1
    height: int = 1


class Grid(FundamentalComponent):
    """
    # Grid

    A container which arranges its children in a table-like grid.

    Grids arrange their children in a table-like grid. Each child is placed in
    one or more cells of the grid. You can add children to the grid either by
    passing them in as a list or by using the `grid_add` method.


    ## Attributes:

    `row_spacing`: The amount of space between rows of the grid.

    `column_spacing`: The amount of space between columns of the grid.


    ## Example:

    This code creates a grid layout with two rows and two columns, and adds
    children to the grid by passing them in as a list:

    ```python
    rio.Grid(
        [rio.Text("Hello"), rio.Text("World!")],  # 1. Row
        [rio.Text("Foo"), rio.Text("Bar")],  # 2. Row
    )
    ```

    Alternatively, you can use the `add_child` method to add children to the grid.
    Here's how you can do it:

    ```python
    grid = rio.Grid(row_spacing=1, column_spacing=1)
    grid.add_child(rio.Text("Hello"), row=0, column=0)
    grid.add_child(rio.Text("World!"), row=0, column=1)
    grid.add_child(rio.Text("Foo"), row=1, column=0)
    grid.add_child(rio.Text("Bar"), row=1, column=1)
    ```

    In your Component class, you can use the `add_child` method within the build function to
    add children to the grid. Here's how you can do it:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            grid = rio.Grid(row_spacing=1, column_spacing=1)
            grid.add_child(rio.Text("Hello"), row=0, column=0)
            grid.add_child(rio.Text("World!"), row=0, column=1)
            grid.add_child(rio.Text("Foo"), row=1, column=0)
            grid.add_child(rio.Text("Bar"), row=1, column=1)

            return grid
    ```
    """

    _: KW_ONLY
    row_spacing: float
    column_spacing: float

    # These must be annotated, otherwise rio won't understand that grids have
    # child components and won't copy over the new values when two Grids are
    # reconciled
    _children: list[rio.Component]
    _child_positions: list[GridChildPosition]

    def __init__(
        self,
        *rows: Iterable[rio.Component],
        row_spacing: float = 0.0,
        column_spacing: float = 0.0,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.row_spacing = row_spacing
        self.column_spacing = column_spacing

        # JS can only work with lists of Components, so we'll store the
        # components and their positions separately
        _children: list[rio.Component] = []
        _child_positions: list[GridChildPosition] = []

        for row_nr, row_components in enumerate(rows):
            for col_nr, component in enumerate(row_components):
                _children.append(component)
                _child_positions.append(
                    GridChildPosition(
                        row=row_nr,
                        column=col_nr,
                    )
                )

        self._children = _children
        self._child_positions = _child_positions

        self._properties_set_by_creator_.update(["_children", "_child_positions"])

    def add_child(
        self,
        child: rio.Component,
        row: int,
        column: int,
        width: int = 1,
        height: int = 1,
    ) -> None:
        """
        Add a child to the grid at a specified position.

        ## Arguments:

        `child`: The child component to add to the grid.

        `row`: The row in which to place the child.

        `column`: The column in which to place the child.

        `width`: The number of columns the child should take up.

        `height`: The number of rows the child should take up.

        ## Example:

        ```python
        grid = rio.Grid(row_spacing=1, column_spacing=1)
        grid.add_child(rio.Text("Hello"), row=0, column=0)
        ```
        """
        if width <= 0:
            raise ValueError("Children have to take up at least one column")

        if height <= 0:
            raise ValueError("Children have to take up at least one row")

        self._children.append(child)
        self._child_positions.append(GridChildPosition(row, column, width, height))

    def _custom_serialize(self) -> JsonDoc:
        return {
            "_children": [child._id for child in self._children],
            "_child_positions": [vars(pos) for pos in self._child_positions],
        }


Grid._unique_id = "Grid-builtin"



from __future__ import annotations

import logging
import tarfile
from pathlib import Path
from typing import *  # type: ignore

from . import common
from .errors import AssetError

_icon_registry: IconRegistry | None = None


class IconRegistry:
    """
    Helper class, which keeps track of all icons known to rio.

    Icons are stored in a .tar.xz archive, and extracted on first use. Any icons which
    have been accessed are kept in memory for rapid access.
    """

    def __init__(self) -> None:
        # Maps icon names (set/icon:variant) to the icon's SVG string. The icon
        # names are canonical form.
        self.cached_icons: dict[str, str] = {}

        # Maps icon set names to the path of the archive file containing the icons
        self.icon_set_archives: dict[str, Path] = {}

    @classmethod
    def get_singleton(cls) -> IconRegistry:
        """
        Return the singleton instance of this class.
        """
        global _icon_registry

        # Create the instance if there is none yet
        if _icon_registry is None:
            _icon_registry = IconRegistry()

            # Register built-in icon sets
            _icon_registry.icon_set_archives["material"] = (
                common.RIO_ASSETS_DIR / "icon-sets" / "material.tar.xz"
            )

            _icon_registry.icon_set_archives["rio"] = (
                common.RIO_ASSETS_DIR / "icon-sets" / "rio.tar.xz"
            )

            _icon_registry.icon_set_archives["styling"] = (
                common.RIO_ASSETS_DIR / "icon-sets" / "styling.tar.xz"
            )

        # Use it
        return _icon_registry

    @staticmethod
    def parse_icon_name(icon_name: str) -> tuple[str, str, str | None]:
        """
        Given a name for an icon, return the three parts of the name: set, icon,
        variant. If the name is syntactically invalid (e.g. too many slashes),
        raise an `AssetError`.
        """
        # Determine the icon set
        sections = icon_name.split("/")

        if len(sections) == 1:
            icon_set = "material"
            icon_name = sections[0]
        elif len(sections) == 2:
            icon_set, icon_name = sections
        else:
            raise AssetError(
                f"Invalid icon name `{icon_name}`. Icons names must be of the form `set/icon:variant`"
            )

        # Determine the icon name and variant
        sections = icon_name.split(":")

        if len(sections) == 1:
            return icon_set, sections[0], None

        if len(sections) == 2:
            return icon_set, sections[0], sections[1]

        raise AssetError(
            f"Invalid icon name `{icon_name}`. Icons names must be of the form `set/icon:variant`"
        )

    @staticmethod
    def normalize_icon_name(icon_name: str) -> str:
        """
        Given a name for an icon, return the canonical form of the name. If the
        name is syntactically invalid (e.g. too many slashes), raise an
        `AssetError`.
        """

        set, name, section = IconRegistry.parse_icon_name(icon_name)

        if section is None:
            return f"{set}/{name}"

        return f"{set}/{name}:{section}"

    def _icon_set_extraction_dir(self, icon_set: str) -> Path:
        """
        Given the name of an icon set, return the directory where the icon set
        will be extracted to. The directory will be created if necessary.
        """
        return common.ASSET_MANGER.get_cache_path(
            Path("icon-sets") / icon_set,
        )

    def _ensure_icon_set_is_extracted(self, icon_set: str) -> None:
        """
        Given the name of an icon set, extract the icon set's archive to the
        cache directory. The target director must not exist yet. Raises a
        `KeyError` if no icon set with the given name has been registered.
        """
        # If the target directory already exists there is nothing to do
        icon_set_dir = self._icon_set_extraction_dir(icon_set)

        if icon_set_dir.exists():
            return

        # Get the path to the icon set's archive. If there is no icon set with
        # the given name, this will raise a `KeyError`. That's fine.
        archive_path = self.icon_set_archives[icon_set]

        # Extract the set
        logging.debug(
            f"Extracting icon set `{icon_set}` from `{archive_path}` to `{icon_set_dir}`"
        )

        with tarfile.open(archive_path, "r:xz") as tar_file:
            tar_file.extractall(icon_set_dir.parent)

        # Sanity check: Make sure the target directory exists now
        if not icon_set_dir.exists():
            raise RuntimeError(
                f"After extracting icon set `{icon_set}` from `{archive_path}` to `{icon_set_dir.parent}`, the target directory does not exist. Is the directory in the archive named correctly?"
            )

    def _get_icon_svg_path(
        self,
        icon_name: str,
    ) -> Path:
        """
        Given an icon name, return the path to the SVG file for that icon. This
        will extract the icon  if necessary.
        """
        # Prepare some paths
        icon_set, icon_name, variant = IconRegistry.parse_icon_name(icon_name)

        icon_set_dir = self._icon_set_extraction_dir(icon_set)

        if variant is None:
            svg_path = icon_set_dir / f"{icon_name}.svg"
        else:
            svg_path = icon_set_dir / variant / f"{icon_name}.svg"

        # Extract the icon set if necessary
        self._ensure_icon_set_is_extracted(icon_set)

        return svg_path

    def get_icon_svg(self, icon_name: str) -> str:
        """
        Given an icon name, return the SVG string for that icon. This will
        extract the icon if necessary. If the icon name is invalid or there is
        no matching icon, raise an `AssetError`.
        """

        # Normalize the icon name
        icon_name = IconRegistry.normalize_icon_name(icon_name)

        # Already cached?
        try:
            return self.cached_icons[icon_name]
        except KeyError:
            pass

        # Get the path to the icon's SVG file
        svg_path = self._get_icon_svg_path(icon_name)

        # Read the SVG file
        try:
            svg_string = svg_path.read_text()
        except FileNotFoundError:
            # Figure out which part of the name is the problem to show a
            # descriptive error message
            icon_set, icon_name, variant = IconRegistry.parse_icon_name(icon_name)

            icon_set_dir = self._icon_set_extraction_dir(icon_set)

            if not icon_set_dir.exists():
                raise AssetError(
                    f"Unknown icon set `{icon_set}`. Known icon sets are: `{'`, `'.join(self.icon_set_archives.keys())}`"
                ) from None

            raise AssetError(
                f"There is no icon named `{icon_name}` in the `{icon_set}` icon set"
            ) from None

        # Cache the icon
        self.cached_icons[icon_name] = svg_string

        # Done
        return svg_string

    def _get_variant_directories(
        self, icon_set: str
    ) -> Iterable[tuple[str | None, Path]]:
        """
        Given the name of an icon set, list the names of all variants in that
        set along with the directory they are stored in.
        """

        # Make sure the icon set is extracted
        self._ensure_icon_set_is_extracted(icon_set)

        # Iterate over all files in the icon set directory. Directories
        # correspond to variants. If icons are found in the root directory, they
        # are part of the default variant.
        icon_set_dir = self._icon_set_extraction_dir(icon_set)
        has_default_variant = False

        for path in icon_set_dir.iterdir():
            if path.is_dir():
                yield (path.name, path)
            else:
                has_default_variant = True

        if has_default_variant:
            yield (None, icon_set_dir)

    def all_icon_sets(self) -> Iterable[str]:
        """
        Return the names of all icon set names known to rio.
        """
        return self.icon_set_archives.keys()

    def all_variants_in_set(self, icon_set: str) -> Iterable[str | None]:
        """
        Given the name of an icon set, list the names of all variants in that
        set.
        """
        for name, _ in self._get_variant_directories(icon_set):
            yield name

    def all_icons_in_set(
        self,
        icon_set: str,
        *,
        variant: str | None = None,
    ) -> Iterable[tuple[str, str | None]]:
        """
        Given the name of an icon set, list all icon names and variants in that
        set. If `variant` is given, only return icons with that variant.
        Otherwise, icons of all variants are returned.

        Raises a `KeyError` if there is not icon set or variant with the given
        name.
        """
        # Find all available variants. This will also extract the icon set if
        # necessary.
        variants = dict(self._get_variant_directories(icon_set))

        # Apply the variant filter. Any `KeyError` is propagated.
        if variant is not None:
            variants = {variant: variants[variant]}

        # Iterate over all variants
        for variant_name, path in variants.items():
            for icon_path in path.iterdir():
                # Only care for SVG files
                if icon_path.is_dir() or icon_path.suffix != ".svg":
                    continue

                # Yield
                icon_name = icon_path.stem
                yield icon_name, variant_name



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import class_container, component

__all__ = [
    "AppRoot",
]


class AppTopBar(component.Component):
    # TODO
    on_press_open: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        thm = self.session.theme

        icons = []
        for icon in ("castle", "error", "archive"):
            icons.append(
                rio.IconButton(
                    icon,
                    style="plain",
                    size=3,
                )
            )

        return class_container.ClassContainer(
            content=rio.Row(
                rio.IconButton(
                    "menu",
                    style="plain",
                    margin_right=1,
                    size=3,
                    on_press=self.on_press_open,
                ),
                rio.Text(
                    self.session.app.name,
                    style=rio.TextStyle(
                        font_size=1.8,
                        fill=thm.primary_palette.foreground,
                    ),
                ),
                rio.Spacer(),
                rio.Row(*icons, spacing=2),
            ),
            classes=["rio-switcheroo-primary"],
            margin_x=1,
            margin_y=0.1,
        )


class Sidebar(component.Component):
    on_press_close: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        return rio.Column(
            rio.Row(
                rio.IconButton(
                    "menu-open",
                    style="plain",
                    size=3,
                    on_press=self.on_press_close,
                ),
                rio.Column(
                    rio.Text(
                        self.session.app.name,
                        style="heading2",
                        align_x=0,
                    ),
                    rio.Text(
                        "TODO: Subtext",
                        style="dim",
                        align_x=0,
                    ),
                    align_y=0,
                ),
                spacing=1,
                margin_x=1,
                margin_y=1,
                align_x=0,
            ),
            rio.SwitcherBar(
                names=["Foo", "Bar"],
                values=["foo", "bar"],
                orientation="vertical",
                width=25,
                margin=2,
                color="primary",
            ),
            rio.Spacer(),
            rio.SwitcherBar(
                names=["Settings"],
                values=["settings"],
                orientation="vertical",
                width=25,
                margin=2,
                color="primary",
            ),
        )


class AppRoot(component.Component):
    _: KW_ONLY
    fallback_build: Callable[[], rio.Component] | None = None

    _sidebar_is_open: bool = False

    def _on_press_open(self) -> None:
        self._sidebar_is_open = True

    def _on_press_close(self) -> None:
        self._sidebar_is_open = False

    def build(self) -> rio.Component:
        thm = self.session.theme

        return rio.Drawer(
            anchor=rio.Stack(
                rio.Rectangle(
                    style=rio.BoxStyle(
                        fill=thm.primary_palette.background,
                    ),
                ),
                rio.Column(
                    AppTopBar(
                        on_press_open=self._on_press_open,
                    ),
                    rio.Card(
                        rio.Stack(
                            rio.PageView(
                                fallback_build=self.fallback_build,
                            ),
                            rio.IconButton(
                                "castle",
                                align_x=1,
                                align_y=1,
                            ),
                        ),
                        corner_radius=(
                            thm.corner_radius_medium,
                            thm.corner_radius_medium,
                            0,
                            0,
                        ),
                        height="grow",
                    ),
                ),
            ),
            content=Sidebar(
                on_press_close=self._on_press_close,
            ),
            is_open=self._sidebar_is_open,
        )



from __future__ import annotations

import collections
import copy
import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import *  # type: ignore

from uniserde import Jsonable, JsonDoc

import rio

from .. import inspection

__all__ = [
    "ClientComponent",
    "ValidationError",
    "Validator",
]


@dataclass
class ClientComponent:
    id: int
    type: str
    state: JsonDoc

    @classmethod
    def from_json(
        cls,
        id: int,
        delta_state: JsonDoc,
        registered_html_components: set[str],
    ) -> "ClientComponent":
        # Don't modify the original dict
        delta_state = copy.deepcopy(delta_state)

        # Get the component type
        try:
            type = delta_state.pop("_type_")
        except KeyError:
            raise ValidationError(f"Component with id `{id}` is missing `_type_` field")

        if not isinstance(type, str):
            raise ValidationError(
                f"Component with id `{id}` has non-string type `{type}`"
            )

        if (
            type
            not in inspection.get_child_component_containing_attribute_names_for_builtin_components()
            and type not in registered_html_components
        ):
            raise ValidationError(f"Component with id `{id}` has unknown type `{type}`")

        # Construct the result
        return cls(
            id=id,
            type=type,
            state=delta_state,
        )

    def _get_child_attribute_names(self) -> Iterable[str]:
        child_attr_names = (
            inspection.get_child_component_containing_attribute_names_for_builtin_components()
        )
        try:
            return child_attr_names[self.type]
        except KeyError:
            return tuple()  # TODO: How to get the children of HTML components?

    @property
    def non_child_containing_properties(
        self,
    ) -> JsonDoc:
        child_attribute_names = self._get_child_attribute_names()

        result = {}
        for name, value in self.state.items():
            if name in child_attribute_names:
                continue

            result[name] = value

        return result

    @property
    def child_containing_properties(
        self,
    ) -> dict[str, None | int | list[int]]:
        child_attribute_names = self._get_child_attribute_names()

        result = {}
        for name, value in self.state.items():
            if name not in child_attribute_names:
                continue

            result[name] = value

        return result

    @property
    def referenced_child_ids(self) -> Iterable[int]:
        for property_value in self.child_containing_properties.values():
            if property_value is None:
                continue

            if isinstance(property_value, int):
                yield property_value
                continue

            assert isinstance(property_value, list), property_value
            yield from property_value

    def __str__(self) -> str:
        # For placeholders, include the python type
        if self.type == "Placeholder":
            component_type = f'{self.type} ({self.state["_python_type_"]})'
        else:
            component_type = self.type

        return f"{component_type} #{self.id}"


class ValidationError(Exception):
    pass


class Validator:
    def __init__(
        self,
        sess: rio.Session,
        *,
        dump_directory_path: Path | None = None,
    ):
        self.session = sess

        if dump_directory_path is not None:
            assert dump_directory_path.exists(), dump_directory_path
            assert dump_directory_path.is_dir(), dump_directory_path

        self.dump_directory_path = dump_directory_path

        self.root_component: ClientComponent | None = None
        self.components_by_id: dict[int, ClientComponent] = {}

        # HTML components must be registered with the frontend before use. This set
        # contains the ids (`FundamentalComponent._unique_id`) of all registered components.
        self.registered_html_components: set[str] = set(
            inspection.get_child_component_containing_attribute_names_for_builtin_components().keys()
        )

    def dump_message(
        self,
        msg: Jsonable,
        *,
        incoming: bool,
    ):
        """
        Dump the message to a JSON file.

        If no path is set in the validator, this function does nothing.
        """
        if self.dump_directory_path is None:
            return

        direction = "incoming" if incoming else "outgoing"
        path = self.dump_directory_path / f"message-{direction}.json"

        with open(path, "w", encoding="utf-8") as f:
            json.dump(msg, f, indent=4)

    def dump_client_state(
        self,
        component: ClientComponent | None = None,
        path: Path | None = None,
    ) -> None:
        """
        Dump the client state to a JSON file.

        If no component is specified, the root component is used.

        If no path is used the Validator's `dump_client_state_path` is used. If
        no path is used and no path set in the validator, this function does
        nothing.
        """
        if path is None and self.dump_directory_path is not None:
            path = self.dump_directory_path / "client-state.json"

        if path is None:
            return

        if component is None:
            assert self.root_component is not None
            component = self.root_component

        with open(path, "w") as f:
            json.dump(
                self.as_json(component),
                f,
                indent=4,
                # The keys are intentionally in a legible order. Don't destroy
                # that.
                sort_keys=False,
            )

    def prune_components(self) -> None:
        """
        Remove all components which are not referenced directly or indirectly by
        the root component.
        """
        # If there is no root component, everybody is an orphan
        if self.root_component is None:
            self.components_by_id.clear()
            return

        # Find all components which are referenced directly or indirectly by the
        # root component
        visited_ids: set[int] = set()

        to_do = [self.root_component]

        while to_do:
            current = to_do.pop()

            # TODO Use this opportunity to detect cycles?
            if current.id in visited_ids:
                print(
                    f"Warning: Validator found a cycle in the component tree involving component with id `{current.id}`"
                )
                continue

            # Mark the current component as visited
            visited_ids.add(current.id)

            # Chain to its children
            for child_id in current.referenced_child_ids:
                to_do.append(self.components_by_id[child_id])

        # Remove all superfluous components
        self.components_by_id = {
            id: component
            for id, component in self.components_by_id.items()
            if id in visited_ids
        }

    def as_json(self, component: ClientComponent | None = None) -> JsonDoc:
        """
        Return a JSON-serializable representation of the client state.
        """

        if component is None:
            assert self.root_component is not None
            component = self.root_component

        result = {
            "_type_": component.type,
            "_id_": component.id,
        }

        for name, value in component.non_child_containing_properties.items():
            result[name] = value

        for name, value in component.child_containing_properties.items():
            if value is None:
                result[name] = None
                continue

            if isinstance(value, int):
                result[name] = self.as_json(self.components_by_id[value])
                continue

            assert isinstance(value, list), value
            result[name] = [self.as_json(self.components_by_id[id]) for id in value]

        return result

    def handle_incoming_message(self, msg: Any) -> None:
        """
        Process a message passed from Client -> Server.

        This will update the `Validator`'s internal client state and validate
        the message, raising a `ValidationError` if any issues are detected.
        """

        # Delegate to the appropriate handler
        try:
            method = msg["method"]
        except KeyError:
            return

        handler_name = f"_handle_incoming_{method}"

        try:
            handler = getattr(self, handler_name)
        except AttributeError:
            return

        handler(msg["params"])

    def handle_outgoing_message(self, msg: Any) -> None:
        """
        Process a message passed from Server -> Client.

        This will update the `Validator`'s internal client state and validate
        the message, raising a `ValidationError` if any issues are detected.
        """

        # Delegate to the appropriate handler
        try:
            method = msg["method"]
        except KeyError:
            return

        handler_name = f"_handle_outgoing_{method}"

        try:
            handler = getattr(self, handler_name)
        except AttributeError:
            return

        handler(msg["params"])

    def _handle_outgoing_updateComponentStates(self, msg: Any) -> None:
        # Dump the message, if requested
        self.dump_message(msg, incoming=False)

        # Update the individual component states
        for component_id, delta_state in msg["deltaStates"].items():
            # Make sure the delta state isn't empty. While this isn't
            # technically invalid, the frontend relies on values such as the
            # margin and alignment to be present. This works, because those
            # values are generated by python regardless of whether they have
            # changed.
            required_keys = {"_type_", "_margin_", "_size_", "_align_", "_grow_"}
            missing_keys = required_keys - set(delta_state.keys())

            if missing_keys:
                delta_state_nice = json.dumps(delta_state, indent=4)
                raise ValidationError(
                    f"Delta state for with id `{component_id}` is missing required keys `{missing_keys}`. The full delta state follows:\n{delta_state_nice}"
                )

            # Get the component's existing state
            try:
                component = self.components_by_id[component_id]
            except KeyError:
                component = ClientComponent.from_json(
                    component_id,
                    delta_state,
                    self.registered_html_components,
                )
                self.components_by_id[component_id] = component
            else:
                delta_state = delta_state.copy()

                # A component's `_type_` cannot be modified. This value is also
                # stored separately by `ClientComponent`, so make sure it never
                # makes it into the component's state.
                try:
                    new_type = delta_state.pop("_type_")
                except KeyError:
                    pass
                else:
                    if new_type != component.type:
                        raise ValidationError(
                            f"Attempted to modify the `_type_` for component with id `{component_id}` from `{component.type}` to `{new_type}`"
                        ) from None

                # Update the component's state
                component.state.update(delta_state)

        # Update the root component if requested
        if msg["rootComponentId"] is not None:
            try:
                self.root_component = self.components_by_id[msg["rootComponentId"]]
            except KeyError:
                raise ValidationError(
                    f"Attempted to set root component to unknown component with id `{msg['rootComponentId']}`"
                ) from None

        # If no root component is known yet, this message has to contain one
        if self.root_component is None:
            raise ValidationError(
                "Despite no root component being known yet, an `UpdateComponentStates` message was sent without a `root_component_id`",
            )

        # Make sure no invalid component references are present
        invalid_references = collections.defaultdict(list)
        for component in self.components_by_id.values():
            for child_id in component.referenced_child_ids:
                if child_id not in self.components_by_id:
                    invalid_references[component.id].append(child_id)

        if invalid_references:
            invalid_references = {
                str(self.components_by_id[component_id]): child_ids
                for component_id, child_ids in invalid_references.items()
            }
            raise ValidationError(
                f"Invalid component references detected: {invalid_references}"
            )

        # Prune the component tree
        self.prune_components()

        # Look for any components which were sent in the message, but are not
        # actually used in the component tree
        ids_sent = set(msg["deltaStates"].keys())
        ids_existing = set(self.components_by_id.keys())
        ids_superfluous = sorted(ids_sent - ids_existing)

        if ids_superfluous:
            print(f"Validator Warning: Message contained superfluous component ids:")

            for id in ids_superfluous:
                delta_state = msg["deltaStates"][id]
                print(
                    f'-  {delta_state.get("_type_", "unknown type")} #{id}  -  {delta_state}'
                )

        # Dump the client state if requested
        self.dump_client_state()

    def _handle_outgoing_evaluateJavascript(self, msg: Any):
        # Is this message registering a new component class?
        match = re.search(r"window.componentClasses\['(.*)'\]", msg["javaScriptSource"])

        if match is None:
            return

        # Remember the component class as registered
        self.registered_html_components.add(match.group(1))



# This file should never be used for anything. It exists, so that VSCode
# recognizes this folder as a Python package. This in turn allows components and
# pages to reference each other, thus enabling refactoring.

from .sample_component import *



from pathlib import Path
from typing import *  # type: ignore

import rio


class AboutMe(rio.Component):
    def build(self) -> rio.Component:
        return rio.Row(
            rio.Image(
                Path("me.jpg"),
                width=11,
                height=11,
                corner_radius=99999,
                fill_mode="zoom",
                margin_right=2,
            ),
            rio.Column(
                rio.Text(
                    "Jane Doe",
                    style="heading1",
                    align_x=0,
                ),
                rio.Text(
                    "Data Analyst",
                    align_x=0,
                ),
                rio.Text(
                    "AI Researcher",
                    align_x=0,
                ),
                rio.Text(
                    "Python Developer",
                    align_x=0,
                ),
                spacing=0.5,
                align_y=0,
                width="grow",
            ),
            width="grow",
        )



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import Any

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["ComponentTree"]


class ComponentTree(FundamentalComponent):
    """
    Note: This component makes not attempt to request the correct amount of
    space. Specify a width/height manually, or make sure it's in a properly
    sized parent.
    """

    _: KW_ONLY

    # Triggered whenever the user selects a component in the tree. The passed
    # value is the component's ID.
    on_select_component: rio.EventHandler[int] = None

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        selected_component_id = msg["selectedComponentId"]

        # Trigger the press event
        await self.call_event_handler(self.on_select_component, selected_component_id)

        # Refresh the session
        await self.session._refresh()


ComponentTree._unique_id = "ComponentTree-builtin"



class NoFileSelectedError(Exception):
    pass


class AssetError(Exception):
    """
    Raised when an error occurs related to assets.

    Signifies that some operation related to assets has failed. E.g. trying to
    access a nonexistent asset.

    Attributes:
        message: A human-readable error message.
    """

    def __init__(self, message: str):
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0]


class NavigationFailed(Exception):
    """
    Raised when navigation to a page fails.

    This exception is raised when attempting to navigate to a page, but the
    navigation fails for some reason. This could happen, for example, because
    a page guard throws an exception.

    Note that navigating to a nonexistent pages is not an error, as `PageViews`
    will simply display their fallback in that case. Thus this exception will
    not be raised in that case.
    """

    pass



import time
from pathlib import Path
from typing import *  # type: ignore

import watchfiles

from .. import project
from . import run_models


class FileWatcherWorker:
    def __init__(
        self,
        *,
        push_event: Callable[[run_models.Event], None],
        proj: project.RioProject,
    ) -> None:
        self.push_event = push_event
        self.proj = proj

    async def run(self) -> None:
        """
        Watch the project directory for changes and report them as events.
        """
        # Only care for certain files
        filter = watchfiles.PythonFilter()

        # Watch the project directory
        async for changes in watchfiles.awatch(
            self.proj.project_directory, watch_filter=filter
        ):
            for change, path in changes:
                path = Path(path)

                # Not all files trigger a reload
                if not self._file_triggers_reload(path):
                    continue

                # Report the change
                self.push_event(
                    run_models.FileChanged(
                        time.monotonic_ns(),
                        path,
                    )
                )

    def _file_triggers_reload(self, path: Path) -> bool:
        """
        Returns True if the given file should trigger a reload of the project.
        """

        if path.suffix == ".py":
            return True

        # TODO: While a change to `rio.toml` should absolutely trigger a reload,
        #   the other code doesn't currently actually reload the project. Hence
        #   it would be confusing to display a reload to the user without
        #   actually doing anything.
        #
        # if path.name in ("rio.toml",):
        #     return True

        return False



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Literal, TypeVar

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Revealer",
    "RevealerChangeEvent",
]

T = TypeVar("T")


@dataclass
class RevealerChangeEvent:
    is_open: bool


class Revealer(FundamentalComponent):
    """
    # Revealer

    A component that can be used to hide and reveal content.

    `Revealer` is a versatile component that can be used to hide and reveal
    content. It can be used to create collapsible sections, or to hide and
    reveal content based on user input.


    ## Attributes:

    `header`: The header of the `Revealer`. If `None`, the `Revealer` will be
            hidden by default.

    `content`: The content to display when the `Revealer` is open.

    `header_style`: The style of the header. Can be one of `"heading1"`,
            `"heading2"`, `"heading3"`, or `"text"`.

    `is_open`: Whether the `Revealer` is open or not.

    `on_change`: An event handler that is called when the `Revealer` is opened
            or closed. The event handler receives a `RevealerChangeEvent` as
            input.


    ## Example:

    A simple `Revealer` that displays "Hello" when opened:

    ```python
    rio.Revealer(header="Click to reveal", content=rio.Text("Hello"))
    ```

    A `Revealer` is a component that hides content, in this case a `TextInput`,
    until it's opened:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def on_change_update_text(self, event: rio.TextInputChangeEvent):
            self.text = event.text

        def build(self) -> rio.Component:
            return rio.Revealer(
                header="Click to Reveal",
                content=rio.TextInput(
                    label="Write your comment here",
                    text=self.text,
                    on_change=self.on_change_update_text,
                ),
                header_style="heading2",
            )
    ```
    """

    header: str | None
    content: rio.Component
    _: KW_ONLY
    header_style: (
        Literal["heading1", "heading2", "heading3", "text"] | rio.TextStyle
    ) = "text"
    is_open: bool = False
    on_change: rio.EventHandler[RevealerChangeEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_open"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_open" in delta_state and self.header is None:
            raise AssertionError(
                f"Frontend tried to set `Revealer.is_open` even though it has no `header`"
            )

    async def _call_event_handlers_for_delta_state(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                RevealerChangeEvent(new_value),
            )

    def _custom_serialize(self) -> JsonDoc:
        # Serialization doesn't handle unions. Hence the custom serialization
        # here
        if isinstance(self.header_style, str):
            header_style = self.header_style
        else:
            header_style = self.header_style._serialize(self.session)

        return {
            "header_style": header_style,
        }


Revealer._unique_id = "Revealer-builtin"



import functools
import json
import re
from dataclasses import dataclass
from pathlib import Path
from typing import *  # type: ignore

from .. import common

SECTION_PATTERN = re.compile(r"#\s*<(\/?[\w-]+)>")

_ALL_SNIPPET_PATHS: dict[str, Path] | None = None


@dataclass
class Snippet:
    # The name the snippet can be accessed via
    name: str

    # Read from the snippet file
    raw_code: str

    # Determined by the snippet's parent directory
    target_directory: Literal["components", "pages", "other"]

    # Read from the snippet's metadata file
    dependencies: set[str]

    def get_section(self, section_name: str) -> str:
        # Find the target section
        lines = self.raw_code.splitlines()
        section_found = False
        result = []

        for line in lines:
            match = SECTION_PATTERN.match(line)

            # No match
            if match is None:
                result.append(line)
                continue

            # Start of the target section?
            if match.group(1) == section_name:
                result = []
                section_found = True
                continue

            # End of the target section?
            if match.group(1) == f"/{section_name}":
                return "\n".join(result)

            # Some other section, drop it
            pass

        # The section was never found
        if not section_found:
            raise KeyError(f'There is no section named "{section_name}"')

        # The section was found, but never closed
        raise ValueError(f'The section "{section_name}" was never closed')

    def stripped_code(self) -> str:
        """
        Returns the given code with all section tags removed.
        """
        lines = self.raw_code.splitlines()

        # Remove all section tags
        result = []
        for line in lines:
            if SECTION_PATTERN.match(line) is None:
                result.append(line)

        return "\n".join(result)


def _scan_snippets() -> None:
    """
    Find all snippets in the snippet directory, and initialize the
    `_ALL_SNIPPET_PATHS` global variable.
    """
    global _ALL_SNIPPET_PATHS
    assert _ALL_SNIPPET_PATHS is None

    # Find all snippet files
    _ALL_SNIPPET_PATHS = {}

    def scan_dir_recursively(group_name: str, path: Path) -> None:
        assert path.is_dir(), path
        assert _ALL_SNIPPET_PATHS is not None

        for fpath in path.iterdir():
            # Directory
            if fpath.is_dir():
                scan_dir_recursively(group_name, fpath)

            # Snippet file
            else:
                key = f"{group_name}/{fpath.name}"
                _ALL_SNIPPET_PATHS[key] = fpath

    # Scan all snippet directories. The first directory is used as a key, the
    # rest just for organization.
    for group_dir in common.SNIPPETS_DIR.iterdir():
        assert group_dir.is_dir(), group_dir
        scan_dir_recursively(group_dir.name, group_dir)


@functools.lru_cache(maxsize=None)
def get_raw_snippet(name: str) -> Snippet:
    """
    Returns the contents of the snippet file with the given name. Raises a
    `KeyError` if no such snippet exists.

    The snippet isn't processed in any way, meaning all tags and metadata are
    still present.
    """
    # Make sure the snippets have been scanned
    if _ALL_SNIPPET_PATHS is None:
        _scan_snippets()

    assert _ALL_SNIPPET_PATHS is not None

    # Read the snippet code
    snippet_path = _ALL_SNIPPET_PATHS[name]
    try:
        raw_code = snippet_path.read_text()
    except KeyError:
        raise KeyError(f'There is no snippet named "{name}"') from None

    # Define the default values for the metadata
    target_directory = snippet_path.parent.name
    if target_directory not in ("components", "pages"):
        target_directory = "other"

    dependencies = set()

    # Try to read the metadata JSON
    json_path = snippet_path.with_suffix(".json")

    try:
        with json_path.open("r") as json_file:
            metadata = json.load(json_file)

    except FileNotFoundError:
        pass

    else:
        dependencies = set(metadata.get("dependencies", dependencies))

    # Construct and return the snippet
    return Snippet(
        name=name,
        raw_code=raw_code,
        target_directory=target_directory,
        dependencies=dependencies,
    )


@functools.lru_cache(maxsize=None)
def get_snippet_section(
    snippet_name: str,
    *,
    section: str | None = None,
) -> str:
    """
    Returns the contents of the specified section in the given snippet. Raises a
    `KeyError` if no such snippet exists.
    """
    snippet = get_raw_snippet(snippet_name)

    # If no section name was given, return the entire snippet
    if section is None:
        return snippet.stripped_code()

    # Otherwise, return the specified section
    return snippet.get_section(section)



from __future__ import annotations

import pathlib
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import common
from .fills import FillLike
from .self_serializing import SelfSerializing

__all__ = [
    "Font",
    "TextStyle",
]


@dataclass(frozen=True)
class Font(SelfSerializing):
    name: str
    regular: pathlib.Path | bytes
    bold: pathlib.Path | bytes | None = None
    italic: pathlib.Path | bytes | None = None
    bold_italic: pathlib.Path | bytes | None = None

    def _serialize(self, sess: rio.Session) -> str:
        sess._register_font(self)
        return self.name


ROBOTO = Font(
    "Roboto",
    regular=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Regular.ttf",
    bold=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Bold.ttf",
    italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-Italic.ttf",
    bold_italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto/Roboto-BoldItalic.ttf",
)

ROBOTO_MONO = Font(
    "Roboto Mono",
    regular=common.HOSTED_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Regular.ttf",
    bold=common.HOSTED_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Bold.ttf",
    italic=common.HOSTED_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Italic.ttf",
    bold_italic=common.HOSTED_ASSETS_DIR
    / "fonts/Roboto Mono/RobotoMono-BoldItalic.ttf",
)


@dataclass(frozen=True)
class TextStyle(SelfSerializing):
    _: KW_ONLY
    font: Font | None = None
    fill: FillLike | None = None
    font_size: float = 1.0
    italic: bool = False
    font_weight: Literal["normal", "bold"] = "normal"
    underlined: bool = False
    all_caps: bool = False

    def replace(
        self,
        *,
        font: Font | None = None,
        fill: FillLike | None = None,
        font_size: float | None = None,
        italic: bool | None = None,
        font_weight: Literal["normal", "bold"] | None = None,
        underlined: bool | None = None,
        all_caps: bool | None = None,
    ) -> TextStyle:
        return type(self)(
            font=self.font if font is None else font,
            fill=self.fill if fill is None else fill,
            font_size=self.font_size if font_size is None else font_size,
            italic=self.italic if italic is None else italic,
            font_weight=self.font_weight if font_weight is None else font_weight,
            underlined=self.underlined if underlined is None else underlined,
            all_caps=self.all_caps if all_caps is None else all_caps,
        )

    def _serialize(self, sess: rio.Session) -> JsonDoc:
        return {
            "fontName": None if self.font is None else self.font._serialize(sess),
            "fill": None if self.fill is None else self.fill._serialize(sess),
            "fontSize": self.font_size,
            "italic": self.italic,
            "fontWeight": self.font_weight,
            "underlined": self.underlined,
            "allCaps": self.all_caps,
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Slider",
    "SliderChangeEvent",
]


@dataclass
class SliderChangeEvent:
    value: float


class Slider(FundamentalComponent):
    """
    # Slider

    A component for selecting a single value from a range.

    The `Slider` components allows the user to select a single real number value
    by dragging a handle along a line. The value can be any number within a
    range you can specify.


    ## Attributes:

    `minimum`: The minimum value the slider can be set to.

    `maximum`: The maximum value the slider can be set to.

    `value`: The current value of the slider.

    `is_sensitive`: Whether the slider should respond to user input.

    `on_change`: A callback that is called when the value of the slider changes.


    ## Example:

    A minimal example of a `Slider` ranging from 0 to 100 will be shown:

    ```python
    rio.Slider(minimum=0, maximum=100)
    ```

    You can easily `bind` state variables to track changes. If you want to make your `Slider` more
    responsive, you can easily achieve this by adding a lambda function call to `on_change`:

    ```python
    class MyComponent(rio.Component):
        value: int = 0

        def build(self) -> rio.Component:
            return rio.Slider(
                value=self.bind().value,  # state binding
                minimum=0,
                maximum=100,
                step_size=1,
                on_change=lambda event: print(f"value: {event.value}"),
            )
    ```

    If you are not using state bindings, you can also use a lambda function for updating the input value:

    ```python
    class MyComponent(rio.Component):
        value: int = 0

        def build(self) -> rio.Component:
            return rio.Slider(
                value=self.value,
                minimum=0,
                maximum=100,
                step_size=1,
                on_change=lambda event: setattr(self, "value", event.value),
            )
    ```

    You can also use a method for updating the input value and do whatever you want. `Note` that methods
    are handy if you want to do more than just updating the input value. For example run async code or
    update other components based on the input text:

    ```python
    class MyComponent(rio.Component):
        value: float = 0

        def on_change(self, event: rio.SliderChangeEvent):
            self.value = event.value
            print(f"value: {self.value}")

        def build(self) -> rio.Component:
            return rio.Slider(
                value=self.value,
                minimum=0,
                maximum=100,
                step_size=1,
                on_change=self.on_change,
            )
    ```
    """

    minimum: float
    maximum: float
    value: float
    step_size: float
    is_sensitive: bool
    on_change: rio.EventHandler[SliderChangeEvent]

    def __init__(
        self,
        *,
        minimum: float = 0,
        maximum: float = 1,
        step_size: float = 0,
        value: float | None = None,
        is_sensitive: bool = True,
        on_change: rio.EventHandler[SliderChangeEvent] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["grow"] = 1.3,
        height: float | Literal["grow"] = 1.3,
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.minimum = minimum
        self.maximum = maximum
        self.step_size = step_size
        self.value = value  # type: ignore  Possibly assigning None. Fixed in __post_init__ below
        self.is_sensitive = is_sensitive
        self.on_change = on_change

    def __post_init__(self) -> None:
        # Don't hammer potential state bindings
        minimum = self.minimum
        maximum = self.maximum
        step_size = self.step_size
        value = self.value

        initial_value = value

        if maximum <= minimum:
            raise ValueError(
                f"`maximum` must be greater than `minimum`. Got {maximum} <= {minimum}"
            )

        if step_size < 0:
            raise ValueError(
                f"`step_size` must be greater than or equal to 0. Got {step_size}"
            )

        if value is None:
            value = minimum + (maximum - minimum) / 2

        if step_size != 0:
            value = round(value / step_size) * step_size

        value = min(maximum, max(minimum, value))

        # Only assign the value if it has in fact changed, as this causes a
        # refresh. If the value is bound to the parent and the parent rebuilds
        # this creates an infinite loop.
        if value != initial_value:
            self.value = value

    # TODO: When `minimum` or `maximum` is changed, make sure the value is still within
    # the range

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"value"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "value" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `Slider.value` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["value"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, (int, float)), new_value
            await self.call_event_handler(
                self.on_change,
                SliderChangeEvent(new_value),
            )


Slider._unique_id = "Slider-builtin"



from __future__ import annotations

import asyncio
import inspect
import io
import sys
import typing
import weakref
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Callable, Iterable
from dataclasses import KW_ONLY, field
from pathlib import Path
from typing import *  # type: ignore

from typing_extensions import Self, dataclass_transform
from uniserde import Jsonable, JsonDoc

import rio

from .. import event, global_state, inspection
from ..dataclass import RioDataclassMeta, class_local_fields, internal_field
from ..state_properties import (
    PleaseTurnThisIntoAStateBinding,
    StateBinding,
    StateBindingMaker,
    StateProperty,
)
from . import fundamental_component

__all__ = ["Component"]


T = TypeVar("T")


async def call_component_handler_once(
    weak_component: weakref.ReferenceType[Component],
    handler: Callable,
) -> None:
    # Does the component still exist?
    component = weak_component()

    if component is None:
        return

    # Call the handler
    await component.call_event_handler(lambda: handler(component))
    await component.session._refresh()


async def _periodic_event_worker(
    weak_component: weakref.ReferenceType[Component],
    handler: Callable,
    period: float,
) -> None:
    try:
        sess = weak_component().session  # type: ignore
    except AttributeError:
        return

    while True:
        # Wait for the next tick
        await asyncio.sleep(period)

        # Wait until there's an active connection to the client. We won't run
        # code periodically if we aren't sure whether the client will come back.
        await sess._is_active_event.wait()

        # Call the handler
        await call_component_handler_once(weak_component, handler)


def _determine_properties_set_by_creator(
    component: Component, args: tuple, kwargs: dict
) -> set[str]:
    # Determine which properties were explicitly set. This includes
    # parameters that received an argument, and also varargs (`*args`)
    signature = inspect.signature(component.__init__)
    bound_args = signature.bind(*args, **kwargs)

    properties_set_by_creator = set(bound_args.arguments)

    # fmt: off
    properties_set_by_creator.update(
        param.name
        for param in signature.parameters.values()
        if param.kind in (
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.VAR_KEYWORD,
        )
    )
    # fmt: on

    # Discard parameters that don't correspond to state properties
    properties_set_by_creator.intersection_update(type(component)._state_properties_)

    return properties_set_by_creator


C = typing.TypeVar("C", bound="Component")


# For some reason vscode doesn't understand that this class is a
# `@dataclass_transform`, so we'll annotate it again...
@dataclass_transform(
    eq_default=False,
    field_specifiers=(internal_field, field),
)
class ComponentMeta(RioDataclassMeta):
    # Cache for the set of all `StateProperty` instances in this class
    _state_properties_: dict[str, StateProperty]

    # Maps event tags to the methods that handle them. The methods aren't bound
    # to the instance yet, so make sure to pass `self` when calling them
    #
    # The assigned value is needed so that the `Component` class itself has a
    # valid value. All subclasses override this value in `__init_subclass__`.
    _rio_event_handlers_: defaultdict[event.EventTag, list[tuple[Callable, Any]]]

    # Whether this component class is built into Rio, rather than user defined,
    # or from a library.
    _rio_builtin_: bool

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Replace all properties with custom state properties
        cls._initialize_state_properties()

        # Inherit event handlers from parents
        cls._rio_event_handlers_ = defaultdict(list)

        for base in cls.__bases__:
            if not isinstance(base, ComponentMeta):
                continue

            for event_tag, handlers in base._rio_event_handlers_.items():  # type: ignore[wtf]
                cls._rio_event_handlers_[event_tag].extend(handlers)

        # Add events from this class itself
        for member in vars(cls).values():
            if not callable(member):
                continue

            try:
                events = member._rio_events_
            except AttributeError:
                continue

            for event_tag, args in events.items():
                for arg in args:
                    cls._rio_event_handlers_[event_tag].append((member, arg))

        # Is this class built into Rio?
        cls._rio_builtin_ = cls.__module__.startswith("rio.")

    def _initialize_state_properties(cls) -> None:
        """
        Spawn `StateProperty` instances for all annotated properties in this
        class.
        """
        all_parent_state_properties: dict[str, StateProperty] = {}

        for base in reversed(cls.__bases__):
            if isinstance(base, ComponentMeta):
                all_parent_state_properties.update(base._state_properties_)  # type: ignore[wtf]

        cls._state_properties_ = all_parent_state_properties

        annotations: dict = vars(cls).get("__annotations__", {})
        module = sys.modules[cls.__module__]

        for field_name, field in class_local_fields(cls).items():
            # Skip internal fields
            if not field.state_property:
                continue

            # Create the StateProperty
            # readonly = introspection.typing.has_annotation(annotation, Readonly
            readonly = False  # TODO

            state_property = StateProperty(
                field_name, readonly, annotations[field_name], module
            )
            setattr(cls, field_name, state_property)

            # Add it to the set of all state properties for rapid lookup
            cls._state_properties_[field_name] = state_property

    def __call__(cls: type[C], *args: object, **kwargs: object) -> C:
        # Inject the session before calling the constructor
        # Fetch the session this component is part of
        if global_state.currently_building_session is None:
            raise RuntimeError(
                "Components can only be created inside of `build` methods."
            )

        component: C = object.__new__(cls)  # type: ignore[wtf]

        session = global_state.currently_building_session
        component._session_ = session

        # Create a unique ID for this component
        component._id = session._next_free_component_id
        session._next_free_component_id += 1

        component._properties_assigned_after_creation_ = set()

        # Call `__init__`
        component.__init__(*args, **kwargs)

        # Some components (like `Grid`) manually mark some properties as
        # explicitly set in their `__init__`, so we must use `.update()` instead
        # of an assignment
        component._properties_set_by_creator_.update(
            _determine_properties_set_by_creator(component, args, kwargs)
        )

        # Store a weak reference to the component's creator
        if global_state.currently_building_component is None:
            component._weak_creator_ = lambda: None
        else:
            component._weak_creator_ = weakref.ref(
                global_state.currently_building_component
            )

        # Keep track of this component's existence
        #
        # Components must be known by their id, so any messages addressed to
        # them can be passed on correctly.
        session._weak_components_by_id[component._id] = component

        session._register_dirty_component(
            component,
            include_children_recursively=False,
        )

        # Some events need attention right after the component is created
        for event_tag, event_handlers in component._rio_event_handlers_.items():
            # Don't register an empty list of handlers, since that would
            # still slow down the session
            if not event_handlers:
                continue

            # Page changes are handled by the session. Register the handler
            if event_tag == event.EventTag.ON_PAGE_CHANGE:
                callbacks = tuple(handler for handler, unused in event_handlers)
                session._page_change_callbacks[component] = callbacks

            # The `periodic` event needs a task to work in
            elif event_tag == event.EventTag.PERIODIC:
                for callback, period in event_handlers:
                    session.create_task(
                        _periodic_event_worker(
                            weakref.ref(component), callback, period
                        ),
                        name=f"`rio.event.periodic` event worker for {component}",
                    )

        # Call `_rio_post_init` for every class in the MRO
        for base in reversed(type(component).__mro__):
            try:
                post_init = vars(base)["_rio_post_init"]
            except KeyError:
                continue

            post_init(component)

        component._properties_assigned_after_creation_.clear()

        return component


class Component(metaclass=ComponentMeta):
    """
    # Component

    Base class for all `rio` components.

    Components are the building blocks of `rio` apps. `rio` ships with many
    useful components out of the box, but you can also subclass a component to
    create your own.


    ## Attributes:

    `key`: A unique identifier for this component. If two components with the
        same key are present during reconciliation they will be considered
        the same component and their state will be preserved. If no key is
        specified, reconciliation falls back to a less precise method, by
        comparing the location of the component in the component tree.

    `margin`: The margin around this component. This is a shorthand for
        setting `margin_left`, `margin_top`, `margin_right` and
        `margin_bottom` to the same value. If multiple conflicting margins
        are specified the most specific one wins. If for example `margin`
        and `margin_left` are both specified, `margin_left` is used for the
        left side, while the other sides use `margin`.

    `margin_x`: The horizontal margin around this component. This is a
        shorthand for setting `margin_left` and `margin_right` to the same
        value. If multiple conflicting margins are specified the most
        specific one wins. If for example `margin_x` and `margin_left` are
        both specified, `margin_left` is used for the left side, while the
        other side uses `margin_x`.

    `margin_y`: The vertical margin around this component. This is a shorthand
        for setting `margin_top` and `margin_bottom` to the same value. If
        multiple conflicting margins are specified the most specific one
        wins. If for example `margin_y` and `margin_top` are both specified,
        `margin_top` is used for the top side, while the other side uses
        `margin_y`.

    `margin_left`: The left margin around this component. If multiple
        conflicting margins are specified this one will be used, since it's
        the most specific. If for example `margin_left` and `margin` are
        both specified, `margin_left` is used for the left side, while the
        other sides use `margin`.

    `margin_top`: The top margin around this component. If multiple
        conflicting margins are specified this one will be used, since it's
        the most specific. If for example `margin_top` and `margin` are both
        specified, `margin_top` is used for the top side, while the other
        sides use `margin`.

    `margin_right`: The right margin around this component. If multiple
        conflicting margins are specified this one will be used, since it's
        the most specific. If for example `margin_right` and `margin` are
        both specified, `margin_right` is used for the right side, while the
        other sides use `margin`.

    `margin_bottom`: The bottom margin around this component. If multiple
        conflicting margins are specified this one will be used, since it's
        the most specific. If for example `margin_bottom` and `margin` are
        both specified, `margin_bottom` is used for the bottom side, while
        the other sides use `margin`.

    `width`: How much horizontal space this component should request during
        layouting. This can be either a number, or one of the special
        values:

    - If `"natural"`, the component will request the minimum amount it
        requires to fit on the screen. For example a `Text` will request
        however much space the characters of that text require. A `Row`
        would request the sum of the widths of its children.

    - If `"grow"`, the component will request all the remaining space in its parent.

    - Please note that the space a `Component` receives during layouting
        may not match the request. As a general rule for example, containers
        try to pass on all available space to children. If you really want a
        `Component` to only take up as much space as requested, consider
        specifying an alignment.

    `height`: How much vertical space this component should request during
        layouting. This can be either a number, or one of the special values:

    - If `"natural"`, the component will request the minimum amount it
        requires to fit on the screen. For example a `Text` will request
        however much space the characters of that text require. A `Row`
        would request the height of its tallest child.

    - If `"grow"`, the component will request all the remaining space in its parent.

    - Please note that the space a `Component` receives during layouting
        may not match the request. As a general rule for example, containers
        try to pass on all available space to children. If you really want a
        `Component` to only take up as much space as requested, consider
        specifying an alignment.

    `align_x`: How this component should be aligned horizontally, if it
        receives more space than it requested. This can be a number between
        0 and 1, where 0 means left-aligned, 0.5 means centered, and 1 means
        right-aligned.

    `align_y`: How this component should be aligned vertically, if it receives
        more space than it requested. This can be a number between 0 and 1,
        where 0 means top-aligned, 0.5 means centered, and 1 means
        bottom-aligned.
    """

    _: KW_ONLY
    key: str | None = internal_field(default=None, init=True)

    width: float | Literal["natural", "grow"] = "natural"
    height: float | Literal["natural", "grow"] = "natural"

    align_x: float | None = None
    align_y: float | None = None

    margin_left: float | None = None
    margin_top: float | None = None
    margin_right: float | None = None
    margin_bottom: float | None = None

    margin_x: float | None = None
    margin_y: float | None = None
    margin: float | None = None

    # TODO: Figure out why `internal_field` isn't working and use it instead
    _id: int = internal_field(init=False)

    # Weak reference to the component's builder. Used to check if the component
    # is still part of the component tree.
    #
    # Dataclasses like to turn this function into a method. Make sure it works
    # both with and without `self`.
    _weak_builder_: Callable[[], Component | None] = internal_field(
        default=lambda *args: None,
        init=False,
    )

    # Weak reference to the component's creator
    _weak_creator_: Callable[[], Component | None] = internal_field(
        init=False,
    )

    # Each time a component is built the build generation in that component's
    # COMPONENT DATA is incremented. If this value no longer matches the value
    # in its builder's COMPONENT DATA, the component is dead.
    _build_generation_: int = internal_field(default=-1, init=False)

    _session_: rio.Session = internal_field(init=False)

    # Remember which properties were explicitly set in the constructor
    _properties_set_by_creator_: set[str] = internal_field(
        init=False, default_factory=set
    )

    # Remember which properties had new values assigned after the component's
    # construction
    _properties_assigned_after_creation_: set[str] = internal_field(init=False)

    # Whether the `on_populate` event has already been triggered for this
    # component
    _on_populate_triggered_: bool = internal_field(default=False, init=False)

    # Whether this instance is internal to Rio, e.g. because the spawning
    # component is a high-level component defined in Rio.
    #
    # In debug mode, this field is initialized by monkeypatches. When running in
    # release mode this value isn't set at all, and the default set below is
    # always used.
    _rio_internal_: bool = internal_field(init=False, default=False)

    # The stackframe which has created this component. Used by the debugger.
    # Only initialized if in debugging mode.
    _creator_stackframe_: tuple[Path, int] = internal_field(init=False)

    @property
    def session(self) -> rio.Session:
        """
        Return the session this component is part of.
        """
        return self._session_

    # There isn't really a good type annotation for this... Self is the closest
    # thing
    def bind(self) -> Self:
        return StateBindingMaker(self)  # type: ignore

    def _custom_serialize(self) -> JsonDoc:
        """
        Return any additional properties to be serialized, which cannot be
        deduced automatically from the type annotations.
        """
        return {}

    @abstractmethod
    def build(self) -> rio.Component:
        """
        Return a component tree which represents the UI of this component.

        Most components define their appearance and behavior by combining other,
        more basic components. This function's purpose is to do exactly that. It
        returns another component (typically a container) which will be
        displayed on the screen.

        The `build` function should be pure, meaning that it does not modify the
        component's state and returns the same result each time it's invoked.
        """
        raise NotImplementedError()

    def _iter_direct_children(self) -> Iterable[Component]:
        for name in inspection.get_child_component_containing_attribute_names(
            type(self)
        ):
            try:
                value = getattr(self, name)
            except AttributeError:
                continue

            if isinstance(value, Component):
                yield value

            if isinstance(value, list):
                value = cast(list[object], value)

                for item in value:
                    if isinstance(item, Component):
                        yield item

    def _iter_direct_and_indirect_child_containing_attributes(
        self,
        *,
        include_self: bool,
        recurse_into_high_level_components: bool,
    ) -> Iterable[Component]:
        # Special case the component itself to handle `include_self`
        if include_self:
            yield self

        if not recurse_into_high_level_components and not isinstance(
            self, fundamental_component.FundamentalComponent
        ):
            return

        # Iteratively yield all children
        to_do = list(self._iter_direct_children())
        while to_do:
            cur = to_do.pop()
            yield cur

            if recurse_into_high_level_components or isinstance(
                cur, fundamental_component.FundamentalComponent
            ):
                to_do.extend(cur._iter_direct_children())

    def _iter_component_tree(self) -> Iterable[Component]:
        """
        Iterate over all components in the component tree, with this component as the root.
        """
        yield self

        if isinstance(self, fundamental_component.FundamentalComponent):
            for child in self._iter_direct_children():
                yield from child._iter_component_tree()
        else:
            build_result = self.session._weak_component_data_by_component[
                self
            ].build_result
            yield from build_result._iter_component_tree()

    async def _on_message(self, msg: Jsonable, /) -> None:
        raise RuntimeError(f"{type(self).__name__} received unexpected message `{msg}`")

    def _is_in_component_tree(self, cache: dict[rio.Component, bool]) -> bool:
        """
        Returns whether this component is directly or indirectly connected to the
        component tree of a session.

        This operation is fast, but has to walk up the component tree to make sure
        the component's parent is also connected. Thus, when checking multiple
        components it can easily happen that the same components are checked over and
        over, resulting on O(n log n) runtime. To avoid this, pass a cache
        dictionary to this function, which will be used to memoize the result.

        Be careful not to reuse the cache if the component hierarchy might have
        changed (for example after an async yield).
        """

        # Already cached?
        try:
            return cache[self]
        except KeyError:
            pass

        # Root component?
        if self is self.session._root_component:
            result = True

        # Has the builder has been garbage collected?
        else:
            builder = self._weak_builder_()
            if builder is None:
                result = False

            # Has the builder since created new build output, and this component
            # isn't part of it anymore?
            else:
                parent_data = self.session._weak_component_data_by_component[builder]
                result = (
                    parent_data.build_generation == self._build_generation_
                    and builder._is_in_component_tree(cache)
                )

        # Cache the result and return
        cache[self] = result
        return result

    @overload
    async def call_event_handler(
        self,
        handler: rio.EventHandler[[]],
    ) -> None:
        ...

    @overload
    async def call_event_handler(
        self,
        handler: rio.EventHandler[[T]],
        event_data: T,
        /,
    ) -> None:
        ...

    async def call_event_handler(
        self,
        handler: rio.EventHandler[...],
        *event_data: object,
    ) -> None:
        """
        Calls an even handler, awaiting it if necessary.

        Call an event handler, if one is present. Await it if necessary. Log and
        discard any exceptions. If `event_data` is present, it will be passed to
        the event handler.
        """
        await self.session._call_event_handler(handler, *event_data)

    async def force_refresh(self) -> None:
        """
        Force a rebuild of this component.

        Most of the time components update automatically when their state
        changes. However, some state mutations are invisible to `Rio`: For
        example, appending items to a list modifies the list, but since no list
        instance was actually assigned to th component, `Rio` will be unaware of
        this change.

        In these cases, you can force a rebuild of the component by calling
        `force_refresh`. This will trigger a rebuild of the component and
        display the updated version on the screen.

        Another common use case is if you wish to update an component while an
        event handler is still running. `Rio` will automatically detect changes
        after event handlers return, but if you are performing a long-running
        operation, you may wish to update the component while the event handler
        is still running. This allows you to e.g. update a progress bar while
        the operation is still running.
        """
        self.session._register_dirty_component(
            self,
            include_children_recursively=False,
        )

        await self.session._refresh()

    def get_debug_details(self) -> dict[str, Any]:
        """
        Used by Rio's debugger to decide which properties to display to a user,
        when they select a component.
        """
        result = {}

        for prop in self._state_properties_:
            # Consider properties starting with an underscore internal
            if prop.startswith("_"):
                continue

            # Keep it
            result[prop] = getattr(self, prop)

        return result

    def __repr__(self) -> str:
        result = f"<{type(self).__name__} id:{self._id}"

        child_strings: list[str] = []
        for child in self._iter_direct_children():
            child_strings.append(f" {type(child).__name__}:{child._id}")

        if child_strings:
            result += " -" + "".join(child_strings)

        return result + ">"

    def _repr_tree_worker(self, file: IO[str], indent: str) -> None:
        file.write(indent)
        file.write(repr(self))

        for child in self._iter_direct_children():
            file.write("\n")
            child._repr_tree_worker(file, indent + "    ")

    def _repr_tree(self) -> str:
        file = io.StringIO()
        self._repr_tree_worker(file, "")
        return file.getvalue()



from __future__ import annotations

from dataclasses import KW_ONLY

import rio

from .. import cursor_style
from .fundamental_component import FundamentalComponent

__all__ = [
    "Rectangle",
]


class Rectangle(FundamentalComponent):
    """
    # Rectangle

    A customizable rectangle shape.

    Rectangles are versatile components that can be used to build up more
    complex elements. While not particularly interesting on their own, combining
    a rectangle with other components allows you to quickly create custom
    buttons, cards, or anything els you may need in your app.

    Rectangles also act as a simple source of animations. They accept two
    styles: A default style for when the user isn't interacting with them, and a
    hover style for when the mouse hovers above them. This, along with their
    `transition_time` attribute allows you to make your app feel dynamic and
    alive.


    ## Attributes:

    `content`: The component to display inside the rectangle.

    `style`: How the rectangle should appear when the user isn't interacting
        with it.

    `hover_style`: The style of the rectangle when the mouse hovers above it.
        If set to `None`, the rectangle will not change its appearance when
        hovered.

    `transition_time`: How many seconds it should take for the rectangle to
        transition between its styles.

    `cursor`: The cursor to display when the mouse hovers above the rectangle.

    `ripple`: Whether to display a Material Design ripple effect when the
        rectangle is hovered or clicked.


    ## Example:

    A minimal example of a rectangle with a text and red background will be shown:

    ```python
    rio.Rectangle(
        content=rio.Text("Hello World!"),
        style=rio.BoxStyle(fill=rio.Color.from_hex("ff0000")),
    )
    ```

    You can fill your `Rectangle` with an image instead of a color:

    ```python
    from pathlib import Path

    PATH = Path(__file__).parent

    rio.Rectangle(
        content=rio.Text("Hello World!"),
        style=rio.BoxStyle(
            fill=rio.ImageFill(
                PATH / "example_image.jpg",
                fill_mode="zoom",
            ),
        ),
    )
    ```
    """

    _: KW_ONLY
    style: rio.BoxStyle
    content: rio.Component | None = None
    hover_style: rio.BoxStyle | None = None
    transition_time: float = 1.0
    cursor: rio.CursorStyle = cursor_style.CursorStyle.DEFAULT
    ripple: bool = False


Rectangle._unique_id = "Rectangle-builtin"



from datetime import timedelta
from pathlib import Path
from typing import *  # type: ignore

import httpx

BASE_URL = "https://rio.dev/api"


class ApiException(Exception):
    """
    Raised when an API request fails.
    """

    def __init__(
        self,
        message: str,
        status_code: int,
    ):
        super().__init__(message, status_code)

    @property
    def message(self) -> str:
        return self.args[0]

    @property
    def status_code(self) -> int:
        return self.args[1]


class RioApi:
    """
    A wrapper around Rio's web API.
    """

    def __init__(
        self,
        access_token: str | None = None,
    ):
        self._access_token = access_token
        self._http_client = httpx.AsyncClient()

    @property
    def is_logged_in(self) -> bool:
        return self._access_token is not None

    async def __aenter__(self) -> "RioApi":
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    async def close(self) -> None:
        """
        Log out, if logged in and close the HTTP client.
        """
        if self.is_logged_in:
            try:
                await self.logout()
            except ApiException:
                pass

        await self._http_client.aclose()

    async def request(
        self,
        endpoint: str,
        *,
        method: Literal["get", "post", "delete"] = "get",
        json: dict[str, Any] | None = None,
        file: BinaryIO | None = None,
    ) -> Any:
        """
        Make a request to the Rio API.
        """
        # Build the request
        headers = {}
        if self._access_token is not None:
            headers["Authorization"] = f"Bearer {self._access_token}"

        if file is not None:
            files = {"file": file}
        else:
            files = None

        # Make the request
        #
        # TODO: Which exception can this throw?
        response = await self._http_client.request(
            method,
            f"{BASE_URL}/{endpoint}",
            headers=headers,
            json=json,
            files=files,
        )

        # Handle errors
        if response.status_code < 200 or response.status_code >= 300:
            raise ApiException(
                message=response.json()["message"],
                status_code=response.status_code,
            )

        # Return the response
        return response.json()

    async def login_with_name_and_password(
        self,
        username_or_email: str,
        password: str,
        *,
        requested_duration: timedelta = timedelta(days=1),
    ) -> None:
        """
        Log in to the Rio API using a username or email and password.

        If the client was already logged in before the old token is forgotten
        without being expired.
        """
        response = await self.request(
            "/auth/createTokenFromNameAndPassword",
            method="post",
            json={
                "usernameOrEmail": username_or_email,
                "password": password,
                "validForSeconds": requested_duration.total_seconds(),
            },
        )

        self._access_token = response["accessToken"]

    async def logout(self) -> None:
        """
        Log out of the Rio API, if logged in. Does nothing if not logged in.
        """
        if self._access_token is None:
            return

        await self.request("/auth/expireToken", method="post")

    async def get_user(self) -> dict[str, Any]:
        """
        Return the user's information, if logged in.
        """
        assert self.is_logged_in, "Must be logged in to get user info"
        return await self.request("/user/me")

    async def create_app(
        self,
        *,
        name: str,
        packed_app: BinaryIO,
        realm: Literal["pro", "free", "test"],
        start: bool,
    ) -> None:
        assert self.is_logged_in, "Must be logged in to create/update an app"

        await self.request(
            "/app/create",
            method="post",
            json={
                "name": name,
                "start": start,
            },
            file=packed_app,
        )



import functools
import inspect
import warnings
from pathlib import Path

import introspection.typing

from .. import global_state
from ..components.component import Component, ComponentMeta
from ..state_properties import PleaseTurnThisIntoAStateBinding, StateProperty

__all__ = [
    "apply_monkeypatches",
]


def apply_monkeypatches() -> None:
    ComponentMeta.__call__ = functools.partialmethod(
        ComponentMeta_call, ComponentMeta.__call__
    )  # type: ignore[wtf]

    StateProperty.__set__ = functools.partialmethod(
        StateProperty_set, StateProperty.__set__
    )  # type: ignore[wtf]


def ComponentMeta_call(cls, wrapped_method, *args, **kwargs):
    component: Component = wrapped_method(cls, *args, **kwargs)

    # Keep track of who created this component
    caller = inspect.stack()[1]
    component._creator_stackframe_ = (Path(caller.filename), caller.lineno)

    # Track whether this instance is internal to Rio. This is the case if
    # this component's creator is defined in Rio.
    creator = global_state.currently_building_component
    if creator is None:
        assert type(component).__qualname__ == "HighLevelRootComponent", type(
            component
        ).__qualname__
        component._rio_internal_ = True
    else:
        component._rio_internal_ = creator._rio_builtin_

    return component


def StateProperty_set(
    self: StateProperty,
    wrapped_method,
    instance: Component,
    value: object,
):
    # Type check the value
    if not isinstance(value, PleaseTurnThisIntoAStateBinding):
        try:
            annotation = self._resolved_annotation
        except AttributeError:
            annotation = introspection.typing.resolve_forward_refs(
                self._raw_annotation,
                self._module,
                mode="ast",
                strict=False,
                treat_name_errors_as_imports=True,
            )
            self._resolved_annotation = annotation

        try:
            valid = introspection.typing.is_instance(
                value, annotation, forward_ref_context=self._module
            )
        except NotImplementedError:
            warnings.warn(
                f"Unable to verify assignment to"
                f" {type(instance).__qualname__}.{self.name} (annotated as"
                f" {self._annotation_as_string()})"
            )
        else:
            if not valid:
                raise TypeError(
                    f"The value {value!r} can't be assigned to"
                    f" {type(instance).__qualname__}.{self.name}, which is"
                    f" annotated as {self._annotation_as_string()}"
                )

    wrapped_method(self, instance, value)



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

import rio

from .component import Component

__all__ = [
    "NumberInput",
    "NumberInputChangeEvent",
    "NumberInputConfirmEvent",
]


# These must be ints so that `integer * multiplier` returns an int and not a
# float
_multiplier_suffixes: Mapping[str, int] = {
    "k": 1_000,
    "m": 1_000_000,
}


@dataclass
class NumberInputChangeEvent:
    value: float


@dataclass
class NumberInputConfirmEvent:
    value: float


class NumberInput(Component):
    """
    # NumberInput

    Like `TextInput`, but specifically for inputting numbers.

    `NumberInput` allows the user to enter a number. This is similar to the
    `TextInput` component, but with some goodies for handling numbers. The value
    is automatically parsed and formatted according to the user's locale, and
    you can specify minimum and maximum values to limit the user's input.

    Number inputs go way beyond just parsing the input using `int` or `float`.
    They try their best to understand what the user is trying to enter, and also
    support suffixes such as "k" and "m" to represent thousands and millions
    respectively.


    ## Attributes:

    `value`: The number currently entered by the user.

    `label`: A short text to display next to the text input.

    `prefix_text`: A short text to display before the text input. Useful for
            displaying currency symbols or other prefixed units.

    `suffix_text`: A short text to display after the text input. Useful for
            displaying currency names or units.

    `minimum`: The minimum value the number can be set to.

    `maximum`: The maximum value the number can be set to.

    `decimals`: The number of decimals to accept. If the user enters more
            decimals, they will be rounded off. If this value is equal to `0`,
            the input's `value` is guaranteed to be an integer, rather than
            float.

    `is_sensitive`: Whether the text input should respond to user input.

    `is_valid`: Visually displays to the user whether the current text is
            valid. You can use this to signal to the user that their input needs
            to be changed.

    `on_change`: Triggered when the user changes the number.

    `on_confirm`: Triggered when the user explicitly confirms their input,
            such as by pressing the "Enter" key.


    ## Example:

    A minimal example of `NumberInput` with a default value of 20.00 will be shown:
    `Note`: The text will not be updated if the user changes the value in the input field.

    ```python
    rio.NumberInput(value=20.00)
    ```

    You can easily `bind` state variables to track changes. If you want to make your `NumberInput` more
    responsive, you can easily achieve this by adding a lambda function call to e.g. `on_change`:

    ```python
    class MyComponent(rio.Component):
        value: float = 0.0

        def build(self) -> rio.Component:
            return rio.NumberInput(
                value=self.bind().value,  # state binding
                label="price",
                prefix_text="$",
                minimum=0,
                decimals=2,
                on_change=lambda event: print(f"value: {event.value}"),
            )
    ```

    If you are not using state bindings, you can also use a lambda function for updating the input value:

     ```python
    class MyComponent(rio.Component):
        value: float = 0.0

        def build(self) -> rio.Component:
            return rio.NumberInput(
                value=self.bind().value,  # Bind the value to the component
                label="price",
                prefix_text="$",
                minimum=0,
                decimals=2,
                on_change=lambda event: setattr(
                    self, "value", event.value
                ),  # Update the value
            )
    ```

    You can also use a method for updating the input value and do whatever you want. `Note` that methods
    are handy if you want to do more than just updating the input value. For example run async code or
    update other components based on the input text:

    ```python
    class MyComponent(rio.Component):
        value: float = 0.0

        def on_change_update_value(self, event: rio.NumberInputChangeEvent):
            self.value = event.value
            print(f"new value: {self.value}")
            # You can do whatever you want with the new value

        def build(self) -> rio.Component:
            return rio.NumberInput(
                value=self.value,
                label="price",
                prefix_text="$",
                minimum=0,
                decimals=2,
                on_change=self.on_change_update_value,
            )
    ```
    """

    value: float = 0
    _: KW_ONLY
    label: str = ""
    prefix_text: str = ""
    suffix_text: str = ""
    minimum: float | None = None
    maximum: float | None = None
    decimals: int = 2
    is_sensitive: bool = True
    is_valid: bool = True
    on_change: rio.EventHandler[NumberInputChangeEvent] = None
    on_confirm: rio.EventHandler[NumberInputConfirmEvent] = None

    def __post_init__(self):
        self._text_input = None

    def _try_set_value(self, raw_value: str) -> bool:
        """
        Parse the given string and update the component's value accordingly.
        Returns `True` if the value was successfully updated, `False` otherwise.
        """

        # Strip the number down as much as possible
        raw_value = raw_value.strip()
        raw_value = raw_value.replace(self.session._thousands_separator, "")

        # If left empty, set the value to 0, if that's allowable
        if not raw_value:
            self.value = 0

            if self.minimum is not None and self.minimum > 0:
                self.value = self.minimum

            if self.maximum is not None and self.maximum < 0:
                self.value = self.maximum

            return True

        # Check for a multiplier suffix
        suffix = raw_value[-1].lower()
        multiplier = 1

        if suffix.isalpha():
            try:
                multiplier = _multiplier_suffixes[suffix]
            except KeyError:
                pass
            else:
                raw_value = raw_value[:-1].rstrip()

        # Try to parse the number
        try:
            value = float(raw_value.replace(self.session._decimal_separator, "."))
        except ValueError:
            self.value = self.value  # Force the old value to stay
            return False

        # Apply the multiplier
        value *= multiplier

        # Limit the number of decimals
        #
        # Ensure the value is an integer, if the number of decimals is 0
        value = round(value, None if self.decimals == 0 else self.decimals)

        # Clamp the value
        minimum = self.minimum
        if minimum is not None:
            value = max(value, minimum)

        maximum = self.maximum
        if maximum is not None:
            value = min(value, maximum)

        # Update the value
        self.value = value
        return True

    async def _on_change(self, ev: rio.TextInputChangeEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_change,
                NumberInputChangeEvent(self.value),
            )

    async def _on_confirm(self, ev: rio.TextInputConfirmEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_confirm,
                NumberInputConfirmEvent(self.value),
            )

    def build(self) -> rio.Component:
        # Format the number
        value_str = f"{self.value:.{self.decimals}f}"
        if self.decimals == 0:
            int_str, frac_str = value_str, ""
        else:
            int_str, frac_str = value_str.split(".")

        # Add thousands separators
        groups = []
        while len(int_str) > 3:
            groups.append(int_str[-3:])
            int_str = int_str[:-3]

        groups.append(int_str)
        int_str = self.session._thousands_separator.join(reversed(groups))

        # Join the strings
        if self.decimals == 0:
            value_str = int_str
        else:
            value_str = int_str + self.session._decimal_separator + frac_str

        # Build the component
        self._text_input = rio.TextInput(
            text=value_str,
            label=self.label,
            prefix_text=self.prefix_text,
            suffix_text=self.suffix_text,
            is_sensitive=self.is_sensitive,
            is_valid=self.is_valid,
            on_change=self._on_change,
            on_confirm=self._on_confirm,
        )
        return self._text_input

    async def grab_keyboard_focus(self) -> None:
        if self._text_input is not None:
            await self._text_input.grab_keyboard_focus()



from pathlib import Path
from typing import *  # type: ignore

import revel
from revel import error, fatal, input, print, success, warning
from typing_extensions import TypeAlias

import rio.cli
import rio.snippets

__all__ = [
    "create_project",
    "TemplatesLiteral",
]

# A literal type that represents the available templates
TemplatesLiteral: TypeAlias = Literal["biography"]

# The available templates, as runtime accessible set
AVAILABLE_TEMPLATES: set[str] = set(get_args(TemplatesLiteral))

# Which snippets must be included for each project template
#
# The first page will be used as the main page, i.e. it will be located at `/`.
# The URLs of other pages are derived from the snippet name.
TEMPLATE_DEPENDENCIES: dict[TemplatesLiteral | None, set[str]] = {
    None: {
        # Components
        "common-templates/sample_component.py",
        # Pages
        "common-templates/sample_page.py",
    },
    "biography": {
        # Assets
        # "tutorial-biography/jane-doe.jpg",
        # "tutorial-biography/john-doe.jpg",
        # Components
        "tutorial-biography/about_me.py",
        "tutorial-biography/contact.py",
        "tutorial-biography/history.py",
        "tutorial-biography/projects.py",
        "tutorial-biography/skill_bars.py",
        # Pages
        "tutorial-biography/biography_page.py",
    },
}


def class_name_from_snippet_name(file_name: str) -> str:
    """
    Given a file name, determine the name of the class that is defined in it.

    e.g. `sample_component.py` -> `SampleComponent`
    """
    assert file_name.endswith(".py")
    assert file_name.count("/") == 1, file_name
    file_name = file_name.split("/")[1][:-3]
    parts = file_name.split("_")
    return "".join(part.capitalize() for part in parts)


def write_init_file(fil: IO, snippets: Iterable[rio.snippets.Snippet]) -> None:
    """
    Write an `__init__.py` file that imports all of the snippets.

    e.g. if told to import snippets `foo.py` and `bar.py`, it will write:

    ```
    from .foo import Foo
    from .bar import Bar
    ```
    """
    for snippet in snippets:
        assert snippet.name.count("/") == 1, snippet.name
        module_name = snippet.name.split("/")[1][:-3]
        class_name = class_name_from_snippet_name(snippet.name)
        fil.write(f"from .{module_name} import {class_name}\n")


def find_all_dependency_snippets(
    base_snippets: Iterable[str],
) -> list[rio.snippets.Snippet]:
    """
    Given a set of snippet names, determine the full set of all snippets that
    must be included to satisfy all dependencies. This includes the base
    snippets as well.
    """
    dependencies: list[rio.snippets.Snippet] = []
    visited_names: set[str] = set(base_snippets)
    to_do: list[str] = list(visited_names)

    while to_do:
        # Fetch the snippet
        snippet_name = to_do.pop()
        snippet = rio.snippets.get_raw_snippet(snippet_name)

        # Add the snippet to the dependencies
        dependencies.append(snippet)

        # Recursively process all of its dependencies
        for child_name in snippet.dependencies:
            if child_name in visited_names:
                continue

            visited_names.add(child_name)
            to_do.append(child_name)

    assert len(visited_names) == len(dependencies), (visited_names, dependencies)
    return dependencies


def generate_root_init(
    fil: TextIO,
    nicename: str,
    project_type: Literal["app", "website"],
    dependencies: Iterable[rio.snippets.Snippet],
    main_page_snippet_name: str,
) -> None:
    # The root component depends on the type of project
    root_component_name = (
        "rio.PageView"  # TODO: Add other components, once they're supported
    )

    # Prepare the different pages
    page_strings = []
    for dep in dependencies:
        if dep.target_directory != "pages":
            continue

        # What's the URL segment for this page?
        if dep.name == main_page_snippet_name:
            url_segment = ""
        else:
            assert dep.name.count("/") == 1, dep.name
            url_segment = dep.name.split("/")[1][:-3]
            url_segment = url_segment.replace("_", "_").lower()

        page_strings.append(
            f"""
        rio.Page(
            page_url={url_segment!r},
            build=pages.{class_name_from_snippet_name(dep.name)},
        ),"""
        )

    page_string = "\n".join(page_strings)

    # Write the result
    fil.write(
        f"""
from typing import *  # type: ignore

import rio

from . import pages
from . import components as comps


# Define a theme for Rio to use.
#
# You can modify the colors here to adapt the appearance of your app or website.
# The most important parameters are listed, but more are available! You can find
# them all in the docs TODO: Add link.
theme = rio.Theme.from_color(
    primary_color=rio.Color.from_hex("b002ef"),
    secondary_color=rio.Color.from_hex("329afc"),
    light=True,
)


# Create the Rio app
app = rio.App(
    name={nicename!r},
    build={root_component_name},
    pages=[{page_string}
    ],
    theme=theme,
)

""".lstrip()
    )

    if project_type == "website":
        fil.write(
            """
# Make the website available as a FastAPI app
#
# This allows you to run it either via the `rio run` command, or using tools
# such as `uvicorn`.
fastapi_app = app.as_fastapi()
"""
        )


def create_project(
    *,
    nicename: str,
    type: Literal["app", "website"],
    template: TemplatesLiteral | None = None,
) -> None:
    """
    Create a new project with the given name. This will directly interact with
    the terminal, asking for input and printing output.
    """

    # Derive the python name from the nicename
    python_name = nicename.strip().replace("-", "_").replace(" ", "_").lower()

    # If the python name is not a valid module name, ask for a different one
    if not python_name.isidentifier():
        print(
            f"`{nicename}` is not a valid Python module name. What should the module be called?"
        )

        def parse_module_name(name: str) -> str:
            if not name.isidentifier():
                raise ValueError()

            return name

        python_name = input("Module name", parse=parse_module_name)

    # Create the target directory
    project_dir = Path.cwd() / nicename
    project_dir.mkdir(parents=True, exist_ok=True)

    # If the project directory already exists it must be empty
    if any(project_dir.iterdir()):
        fatal(f"The project directory `{project_dir}` already exists and is not empty")

    # Create the config file
    with open(project_dir / "rio.toml", "w") as f:
        f.write("# This is the configuration file for Rio,\n")
        f.write("# an easy to use app & web framework for Python.\n")
        f.write("\n")
        f.write(f"[app]\n")
        f.write(f'app_type = "{type}"  # This is either "website" or "app"\n')
        f.write(f'main_module = "{python_name}"  # The name of your Python module\n')

    # Create the main module and its subdirectories
    main_module_dir = project_dir / python_name
    main_module_dir.mkdir()
    (main_module_dir / "components").mkdir()
    (main_module_dir / "pages").mkdir()
    (main_module_dir / "assets").mkdir()

    # Determine all components and pages that should be copied over
    dependencies: list[rio.snippets.Snippet] = find_all_dependency_snippets(
        TEMPLATE_DEPENDENCIES[template]
    )

    # Copy over all of the dependencies
    for snippet in dependencies:
        assert snippet.target_directory in ("components", "pages"), (
            snippet.name,
            snippet.target_directory,
        )

        assert snippet.name.count("/") == 1, snippet.name
        target_filename = snippet.name.split("/")[1]
        target_path = main_module_dir / snippet.target_directory / target_filename

        target_path.write_text(snippet.stripped_code())

    # Find the main page
    for main_page_snippet in dependencies:
        if main_page_snippet.target_directory == "pages":
            break
    else:
        raise AssertionError(f"Template `{template}` has no main page!?")

    # Generate /project/__init__.py
    with open(main_module_dir / "__init__.py", "w") as fil:
        generate_root_init(
            fil=fil,
            nicename=nicename,
            project_type=type,
            dependencies=dependencies,
            main_page_snippet_name=main_page_snippet.name,
        )

    # Generate /project/pages/__init__.py
    with open(main_module_dir / "pages" / "__init__.py", "w") as f:
        write_init_file(
            f,
            [snip for snip in dependencies if snip.target_directory == "pages"],
        )

    # Generate /project/components/__init__.py
    with open(main_module_dir / "components" / "__init__.py", "w") as f:
        write_init_file(
            f,
            [snip for snip in dependencies if snip.target_directory == "components"],
        )

    # Generate README.md
    #
    # This file isn't strictly necessary, but `poetry install` in particular
    # fails without. So hey, why not make a template.
    with open(project_dir / "README.md", "w") as f:
        f.write(
            f"""# {nicename}

This is a placeholder README for your project. Use it to describe what your
project is about, to give new users a quick overview of what they can expect.

_{nicename.capitalize()} was created using [Rio](http://rio.dev/), an easy to use
app & website framework for Python._
"""
        )

    # Applications require a `__main__.py` as well
    if type == "app":
        with open(main_module_dir / "__main__.py", "w") as f:
            f.write(
                f"""
# Make sure the project is in the Python path
import sys
from pathlib import Path
sys.path.insert(0, str(Path(__file__).parent.parent))


# Import the main module
import {python_name}

# Run the app
{python_name}.app.run_in_window()
"""
            )

    # Report success
    #
    # TODO: Add a command to install dependencies? Activate the venv?
    print()
    success(f"The project has been created!")
    success(f"You can find it at `{project_dir.resolve()}`")
    print()
    print(f"To see your new project in action, run the following commands:")
    print(f"> cd {revel.shell_escape(project_dir.resolve())}")
    print(f"> rio run")



from __future__ import annotations

import functools
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from typing_extensions import Self
from uniserde import Jsonable

import rio

from . import color
from .text_style import ROBOTO, ROBOTO_MONO

# This module imports `curses` even though it doesn't need it, which causes
# problems on Windows. Since it's unused, we'll just give it a fake module if
# the real curses isn't available.
try:
    import material_color_utilities_python
except ImportError:
    import sys
    import types

    sys.modules["curses"] = types.SimpleNamespace(termattrs=None)  # type: ignore
    import material_color_utilities_python

    del sys.modules["curses"]


__all__ = [
    "Palette",
    "Theme",
]


T = TypeVar("T")


def elvis(optional: Optional[T], default: T) -> T:
    if optional is None:
        return default
    else:
        return optional


class MaterialPalette:
    def __init__(self, hue: int, chroma: int):
        self._material_palette = material_color_utilities_python.TonalPalette(
            hue, chroma
        )

    def __getitem__(self, tone: int) -> rio.Color:
        return rio.Color._from_material_argb(self._material_palette.tone(tone))


@dataclass(frozen=True)
class Palette:
    background: rio.Color
    background_variant: rio.Color
    background_active: rio.Color

    foreground: rio.Color

    @classmethod
    def from_color(
        cls,
        color: rio.Color,
        is_light_theme: bool,
    ) -> Self:
        as_hct = material_color_utilities_python.Hct.fromInt(color._as_material_argb)
        hct_palette = MaterialPalette(as_hct.hue, as_hct.chroma)

        return cls(
            background=color,
            background_variant=hct_palette[as_hct.tone + 5],
            background_active=hct_palette[as_hct.tone + 15],
            foreground=hct_palette[8 if as_hct.tone > 60 else 92],
        )

    def replace(
        self,
        background: rio.Color | None = None,
        background_variant: rio.Color | None = None,
        background_active: rio.Color | None = None,
        foreground: rio.Color | None = None,
    ) -> Palette:
        return Palette(
            background=elvis(background, self.background),
            background_variant=elvis(background_variant, self.background_variant),
            background_active=elvis(background_active, self.background_active),
            foreground=elvis(foreground, self.foreground),
        )


@dataclass(frozen=True)
class Theme:
    """
    Defines the visual style of the application.

    The `Theme` contains all colors, text styles, and other visual properties
    that are used throughout the application. If you wish to change the
    appearance of your app, this is the place to do it.

    TODO: Finalize theming and document it

    TODO: Give an example for how to create a theme and attach it to the
        session.
    """

    _: KW_ONLY

    primary_palette: Palette
    secondary_palette: Palette

    background_palette: Palette
    neutral_palette: Palette
    hud_palette: Palette
    disabled_palette: Palette

    success_palette: Palette
    warning_palette: Palette
    danger_palette: Palette

    # Other
    corner_radius_small: float
    corner_radius_medium: float
    corner_radius_large: float

    shadow_color: rio.Color

    font: rio.Font
    monospace_font: rio.Font

    # Text styles
    heading1_style: rio.TextStyle
    heading2_style: rio.TextStyle
    heading3_style: rio.TextStyle
    text_style: rio.TextStyle

    @classmethod
    def from_color(
        cls,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.5,
        corner_radius_medium: float = 1.4,
        corner_radius_large: float = 2.4,
        color_headings: bool | Literal["auto"] = "auto",
        font: rio.Font = ROBOTO,
        monospace_font: rio.Font = ROBOTO_MONO,
        light: bool = True,
    ) -> Self:
        # Impute defaults
        if primary_color is None:
            primary_color = rio.Color.from_hex("b002ef")

        if secondary_color is None:
            secondary_color = rio.Color.from_hex("329afc")

        if success_color is None:
            success_color = rio.Color.from_hex("1E8E3E")

        if warning_color is None:
            warning_color = rio.Color.from_hex("F9A825")

        if danger_color is None:
            danger_color = rio.Color.from_hex("B3261E")

        # Extract palettes from the material theme
        primary_palette = Palette.from_color(primary_color, light)
        secondary_palette = Palette.from_color(secondary_color, light)

        if light:
            if background_color is None:
                background_palette = Palette(
                    background=rio.Color.WHITE,
                    background_variant=rio.Color.from_grey(0.96).blend(
                        primary_color, 0.04
                    ),
                    background_active=rio.Color.from_grey(0.96).blend(
                        primary_color, 0.1
                    ),
                    foreground=rio.Color.from_grey(0.15),
                )
            else:
                background_palette = Palette.from_color(background_color, light)

            if neutral_color is None:
                neutral_palette = Palette(
                    background=rio.Color.from_grey(0.97).blend(primary_color, 0.04),
                    background_variant=rio.Color.from_grey(0.93).blend(
                        primary_color, 0.07
                    ),
                    background_active=rio.Color.from_grey(0.93).blend(
                        primary_color, 0.15
                    ),
                    foreground=rio.Color.from_grey(0.1),
                )
            else:
                neutral_palette = Palette.from_color(neutral_color, light)

            if hud_color is None:
                hud_palette = Palette.from_color(
                    rio.Color.from_grey(
                        0.06,
                        opacity=0.9,
                    ),
                    light,
                )
            else:
                hud_palette = Palette.from_color(hud_color, light)

            if disabled_color is None:
                disabled_palette = Palette(
                    rio.Color.from_grey(0.7),
                    rio.Color.from_grey(0.75),
                    rio.Color.from_grey(0.80),
                    rio.Color.from_grey(0.4),
                )
            else:
                disabled_palette = Palette.from_color(disabled_color, light)

            shadow_color = rio.Color.from_rgb(0.1, 0.1, 0.4, 0.3)

        else:
            if background_color is None:
                background_palette = Palette(
                    background=rio.Color.from_grey(0.08).blend(primary_color, 0.02),
                    background_variant=rio.Color.from_grey(0.14).blend(
                        primary_color, 0.04
                    ),
                    background_active=rio.Color.from_grey(0.14).blend(
                        primary_color, 0.10
                    ),
                    foreground=rio.Color.from_grey(0.9),
                )
            else:
                background_palette = Palette.from_color(background_color, light)

            if neutral_color is None:
                neutral_palette = Palette(
                    background=rio.Color.from_grey(0.16).blend(primary_color, 0.03),
                    background_variant=rio.Color.from_grey(0.2).blend(
                        primary_color, 0.04
                    ),
                    background_active=rio.Color.from_grey(0.2).blend(
                        primary_color, 0.10
                    ),
                    foreground=rio.Color.from_grey(0.5),
                )
            else:
                neutral_palette = Palette.from_color(neutral_color, light)

            if hud_color is None:
                hud_palette = Palette.from_color(
                    rio.Color.from_grey(
                        0.2,
                        opacity=0.8,
                    ),
                    light,
                )
            else:
                hud_palette = Palette.from_color(hud_color, light)

            if disabled_color is None:
                disabled_palette = Palette(
                    rio.Color.from_grey(0.2),
                    rio.Color.from_grey(0.15),
                    rio.Color.from_grey(0.10),
                    rio.Color.from_grey(0.6),
                )
            else:
                disabled_palette = Palette.from_color(disabled_color, light)

            shadow_color = rio.Color.from_rgb(0.0, 0.0, 0.1, 0.35)

        # Semantic colors
        success_palette = Palette.from_color(success_color, light)
        warning_palette = Palette.from_color(warning_color, light)
        danger_palette = Palette.from_color(danger_color, light)

        # Colorful headings can be a problem when the primary color is similar
        # to the background/neutral color. If the `color_headings` argument is
        # set to `auto`, disable coloring if the colors are close.
        if color_headings == "auto":
            brightness1 = primary_palette.background.perceived_brightness
            brightness2 = background_palette.background.perceived_brightness

            color_headings = abs(brightness1 - brightness2) > 0.3

        # Text styles
        text_color = rio.Color.from_grey(0.1 if light else 0.9)

        heading1_style = rio.TextStyle(
            font_size=3.0,
            fill=primary_color if color_headings else text_color,
        )
        heading2_style = heading1_style.replace(font_size=1.8)
        heading3_style = heading1_style.replace(font_size=1.2)
        text_style = heading1_style.replace(
            font_size=1,
            fill=text_color,
        )

        return cls(
            primary_palette=primary_palette,
            secondary_palette=secondary_palette,
            background_palette=background_palette,
            neutral_palette=neutral_palette,
            hud_palette=hud_palette,
            disabled_palette=disabled_palette,
            success_palette=success_palette,
            warning_palette=warning_palette,
            danger_palette=danger_palette,
            corner_radius_small=corner_radius_small,
            corner_radius_medium=corner_radius_medium,
            corner_radius_large=corner_radius_large,
            shadow_color=shadow_color,
            font=font,
            monospace_font=monospace_font,
            heading1_style=heading1_style,
            heading2_style=heading2_style,
            heading3_style=heading3_style,
            text_style=text_style,
        )

    @classmethod
    def pair_from_color(
        cls,
        *,
        primary_color: rio.Color | None = None,
        secondary_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        neutral_color: rio.Color | None = None,
        hud_color: rio.Color | None = None,
        disabled_color: rio.Color | None = None,
        success_color: rio.Color | None = None,
        warning_color: rio.Color | None = None,
        danger_color: rio.Color | None = None,
        corner_radius_small: float = 0.6,
        corner_radius_medium: float = 1.6,
        corner_radius_large: float = 2.6,
        font: rio.Font = ROBOTO,
        monospace_font: rio.Font = ROBOTO_MONO,
        color_headings: bool | Literal["auto"] = "auto",
    ) -> tuple[Self, Self]:
        func = functools.partial(
            cls.from_color,
            primary_color=primary_color,
            secondary_color=secondary_color,
            background_color=background_color,
            neutral_color=neutral_color,
            hud_color=hud_color,
            disabled_color=disabled_color,
            success_color=success_color,
            warning_color=warning_color,
            danger_color=danger_color,
            corner_radius_small=corner_radius_small,
            corner_radius_medium=corner_radius_medium,
            corner_radius_large=corner_radius_large,
            font=font,
            monospace_font=monospace_font,
            color_headings=color_headings,
        )
        return (
            func(light=True),
            func(light=False),
        )

    def text_color_for(self, color: rio.Color) -> rio.Color:
        """
        Given the color of a background, return which color should be used for
        text on top of it.
        """
        if color.perceived_brightness > 0.6:
            return rio.Color.from_grey(0.1)
        else:
            return rio.Color.from_grey(0.9)

    def _serialize_colorset(self, color: color.ColorSet) -> Jsonable:
        # If the color is a string, just pass it through
        if isinstance(color, str):
            return color

        # If it is a custom color, return it, along with related ones
        return {
            "background": color.rgba,
            "foreground": self.text_color_for(color).rgba,
        }

    def replace(
        self,
        primary_palette: Palette | None = None,
        secondary_palette: Palette | None = None,
        background_palette: Palette | None = None,
        neutral_palette: Palette | None = None,
        hud_palette: Palette | None = None,
        disabled_palette: Palette | None = None,
        success_palette: Palette | None = None,
        warning_palette: Palette | None = None,
        danger_palette: Palette | None = None,
        corner_radius_small: float | None = None,
        corner_radius_medium: float | None = None,
        corner_radius_large: float | None = None,
        shadow_color: rio.Color | None = None,
        font_family: rio.Font | None = None,
        monospace_font: rio.Font | None = None,
        heading1_style: rio.TextStyle | None = None,
        heading2_style: rio.TextStyle | None = None,
        heading3_style: rio.TextStyle | None = None,
        text_style: rio.TextStyle | None = None,
    ) -> Theme:
        return Theme(
            primary_palette=elvis(primary_palette, self.primary_palette),
            secondary_palette=elvis(secondary_palette, self.secondary_palette),
            background_palette=elvis(background_palette, self.background_palette),
            neutral_palette=elvis(neutral_palette, self.neutral_palette),
            hud_palette=elvis(hud_palette, self.hud_palette),
            disabled_palette=elvis(disabled_palette, self.disabled_palette),
            success_palette=elvis(success_palette, self.success_palette),
            warning_palette=elvis(warning_palette, self.warning_palette),
            danger_palette=elvis(danger_palette, self.danger_palette),
            corner_radius_small=elvis(corner_radius_small, self.corner_radius_small),
            corner_radius_medium=elvis(corner_radius_medium, self.corner_radius_medium),
            corner_radius_large=elvis(corner_radius_large, self.corner_radius_large),
            shadow_color=elvis(shadow_color, self.shadow_color),
            font=elvis(font_family, self.font),
            monospace_font=elvis(monospace_font, self.monospace_font),
            heading1_style=elvis(heading1_style, self.heading1_style),
            heading2_style=elvis(heading2_style, self.heading2_style),
            heading3_style=elvis(heading3_style, self.heading3_style),
            text_style=elvis(text_style, self.text_style),
        )

    @property
    def is_light_theme(self) -> bool:
        return self.primary_palette.background.perceived_brightness >= 0.5

    @property
    def primary_color(self) -> rio.Color:
        return self.primary_palette.background

    @property
    def secondary_color(self) -> rio.Color:
        return self.secondary_palette.background

    @property
    def background_color(self) -> rio.Color:
        return self.background_palette.background

    @property
    def neutral_color(self) -> rio.Color:
        return self.neutral_palette.background

    @property
    def hud_color(self) -> rio.Color:
        return self.hud_palette.background

    @property
    def disabled_color(self) -> rio.Color:
        return self.disabled_palette.background

    @property
    def success_color(self) -> rio.Color:
        return self.success_palette.background

    @property
    def warning_color(self) -> rio.Color:
        return self.warning_palette.background

    @property
    def danger_color(self) -> rio.Color:
        return self.danger_palette.background



from __future__ import annotations

from typing import Literal

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["Stack"]


class Stack(FundamentalComponent):
    """
    # Stack

    A container that stacks its children in the Z direction.

    `Stacks` are similar to rows and columns, but they stack their children in
    the Z direction instead of the X or Y direction. In other words, the stack's
    children overlap each other, with the first one at the bottom, the second
    one above that, and so on.


    ## Attributes:

    `children`: The components to place in this `Stack`.

    ## Example:

    A minimal example of `Stack` will be shown. This example will create a stack
    of three icons, each with a different size and color. The first icon will be
    at the bottom, the second one above that, and the third one at the top:

    ```python
    rio.Stack(
        # bottom
        rio.Icon(
            "castle",
            width=50,
            height=50,
            fill=rio.Color.from_hex("00ff00"),
        ),
        # middle
        rio.Icon(
            "castle",
            width=30,
            height=30,
            fill=rio.Color.from_hex("ff0000"),
        ),
        # top
        rio.Icon(
            "castle",
            width=10,
            height=10,
            fill=rio.Color.from_hex("000000"),
        ),
    )
    ```
    """

    children: list[rio.Component]

    def __init__(
        self,
        *children: rio.Component,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)


Stack._unique_id = "Stack-builtin"



"""
Pretty-strings a traceback. The result looks very similar to Python's default,
but is colored and just tweaked in general.
"""

import html
import io
import linecache
import sys
import traceback
from pathlib import Path
from typing import *  # type: ignore

import revel


def _format_single_exception_raw(
    out: IO[str],
    err: BaseException,
    *,
    include_header: bool,
    escape: Callable[[str], str],
    bold: str,
    nobold: str,
    dim: str,
    nodim: str,
    yellow: str,
    noyellow: str,
    red: str,
    nored: str,
    relpath: Path | None = None,
    frame_filter: Callable[[traceback.FrameSummary], bool],
) -> None:
    # Get the traceback as a list of frames
    tb_list = traceback.extract_tb(err.__traceback__)

    # Syntax errors are very special snowflakes and need separate treatment.
    #
    # The exact arguments depend on the python version.
    #
    # TODO: Is there a better way to do this? Since Python obviously isn't
    # keeping the arguments consistent, this could potentially break every
    # single Python version.
    if isinstance(err, SyntaxError):
        if sys.version_info < (3, 11):
            frame = traceback.FrameSummary(
                filename=err.filename,  # type: ignore
                lineno=err.lineno,
                name="<module>",
                line=err.text,
                locals=None,
            )
        else:
            frame = traceback.FrameSummary(
                filename=err.filename,  # type: ignore
                lineno=err.lineno,
                end_lineno=err.end_lineno,
                colno=err.offset,
                end_colno=err.end_offset,
                name="<module>",
                line=err.text,
                locals=None,
            )

        tb_list.append(frame)

    # Lead-in
    if include_header:
        out.write(f"{dim}Traceback (most recent call last):{nodim}\n")

    # Iterate through frames
    for frame in tb_list:
        # Keep this frame?
        if not frame_filter(frame):
            continue

        # Make paths relative to the relpath if they're inside it
        frame_path = Path(frame.filename)

        if relpath is not None and frame_path.is_absolute():
            try:
                frame_path = frame_path.relative_to(relpath)
            except ValueError:
                pass

        # Format the file location
        out.write(
            f"  {dim}File{nodim} {yellow}{escape(str(frame_path))}{noyellow}{dim}, {nodim}{yellow}line {frame.lineno}{noyellow}{dim}, in {escape(frame.name)}{nodim}\n"
        )

        # Display the source code from that line
        #
        # Insanely, the line in the frame has been stripped. Fetch it again.
        assert frame.lineno is not None  # Doesn't seem to happen, ever
        line = linecache.getline(frame.filename, frame.lineno)

        if line.strip():
            # If this is the last line, highlight the error
            if (
                frame is tb_list[-1]
                and hasattr(frame, "colno")
                and hasattr(frame, "end_colno")
            ):
                assert frame.colno is not None  # type: ignore
                assert frame.end_colno is not None  # type: ignore

                before = escape(line[: frame.colno])  # type: ignore
                error = escape(line[frame.colno : frame.end_colno])  # type: ignore
                after = escape(line[frame.end_colno :])  # type: ignore
                line = f"{before}{red}{error}{nored}{after}"
            else:
                line = escape(line)

            # NOW strip it
            line = line.strip()
            out.write(f"    {line}\n")

    # Actual error message
    out.write("\n")
    out.write(f"{bold}{red}{type(err).__name__}{nored}{nobold}")

    error_message = str(err)
    if error_message:
        out.write(f"{bold}: {escape(str(err))}{nobold}")


def format_exception_raw(
    err: BaseException,
    *,
    escape: Callable[[str], str],
    bold: str,
    nobold: str,
    dim: str,
    nodim: str,
    yellow: str,
    noyellow: str,
    red: str,
    nored: str,
    relpath: Path | None = None,
    frame_filter: Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    def format_inner(err: BaseException) -> None:
        # Chain to the cause or context if there is one
        if err.__cause__ is not None:
            format_inner(err.__cause__)
            out.write("\n\n")
            out.write(f"The above exception was the direct cause of the following:\n\n")
            include_header = False

        elif err.__context__ is not None:
            format_inner(err.__context__)
            out.write("\n\n")
            out.write(
                f"During handling of the above exception, another exception occurred:\n\n"
            )
            include_header = False
        else:
            include_header = True

        # Format this exception
        _format_single_exception_raw(
            out,
            err,
            include_header=include_header,
            escape=escape,
            bold=bold,
            nobold=nobold,
            dim=dim,
            nodim=nodim,
            yellow=yellow,
            noyellow=noyellow,
            red=red,
            nored=nored,
            relpath=relpath,
            frame_filter=frame_filter,
        )

    # Start the recursion
    out = io.StringIO()
    format_inner(err)

    # Return the result
    return out.getvalue()


def format_exception_revel(
    err: BaseException,
    *,
    relpath: Path | None = None,
    frame_filter: Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    return format_exception_raw(
        err,
        bold="[bold]",
        nobold="[/]",
        dim="[dim]",
        nodim="[/]",
        yellow="[yellow]",
        noyellow="[/]",
        red="[red]",
        nored="[/]",
        escape=revel.escape,
        relpath=relpath,
        frame_filter=frame_filter,
    )


def format_exception_html(
    err: BaseException,
    *,
    relpath: Path | None = None,
    frame_filter: Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    result = format_exception_raw(
        err,
        bold='<span class="rio-traceback-bold">',
        nobold="</span>",
        dim='<span class="rio-traceback-dim">',
        nodim="</span>",
        yellow='<span class="rio-traceback-yellow">',
        noyellow="</span>",
        red='<span class="rio-traceback-red">',
        nored="</span>",
        escape=html.escape,
        relpath=relpath,
        frame_filter=frame_filter,
    )

    return result.replace("\n", "<br>")



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import Literal

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["FlowContainer"]


class FlowContainer(FundamentalComponent):
    """
    # FlowContainer

    A container that lays out its children in a horizontal or vertical flow.

    `FlowContainer` is a container that lays out its children in a horizontal or
    vertical flow. It is similar to `Container`, but allows you to specify
    spacing between the children.


    ## Attributes:

    `children`: The components to place inside the container.

    `spacing_x`: The horizontal spacing between the children.

    `spacing_y`: The vertical spacing between the children.


    ## Example:

    This minimal example will show a container with a horizontal flow:

    ```python
    rio.FlowContainer(
        rio.Text("Hello"),
        rio.Text("World!"),
        spacing_x=1,
    )
    ```
    """

    children: list[rio.Component]
    _: KW_ONLY
    spacing_x: float
    spacing_y: float

    def __init__(
        self,
        *children: rio.Component,
        spacing_x: float = 0.0,
        spacing_y: float = 0.0,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        assert isinstance(children, tuple), children
        for child in children:
            assert isinstance(child, rio.Component), child

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.spacing_x = spacing_x
        self.spacing_y = spacing_y


FlowContainer._unique_id = "FlowContainer-builtin"



from __future__ import annotations

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "CodeExplorer",
]


class CodeExplorer(FundamentalComponent):
    # TODO
    source_code: str
    build_result: rio.Component

    line_indices_to_component_keys: list[str | None]


CodeExplorer._unique_id = "CodeExplorer-builtin"



# This module should only contain fonts. The `Font` class is purposely defined
# somewhere else. We don't want `Font` to pop up when someone types `rio.font.`
# in their IDE.

from .text_style import ROBOTO, ROBOTO_MONO

__all__ = ["ROBOTO", "ROBOTO_MONO"]



from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio


class Contact(rio.Component):
    text: str
    icon_name_or_url: str | rio.URL
    _: KW_ONLY
    link: rio.URL | None = None

    def build(self) -> rio.Component:
        # Prepare the image. This varies depending on whether an icon or an
        # image should be displayed.
        if isinstance(self.icon_name_or_url, str):
            image = rio.Icon(
                self.icon_name_or_url,
                width=2,
                height=2,
                margin_x=0.5,
            )
        else:
            image = rio.Image(
                self.icon_name_or_url,
                width=2,
                height=2,
                margin_x=0.5,
            )

        # Combine everything
        result = rio.Row(
            image,
            rio.Text(
                self.text,
                align_x=0,
                width="grow",
            ),
            width="grow",
            margin=0.2,
        )

        # If a link is provided, wrap the result in a link
        if self.link is not None:
            result = rio.Link(
                result,
                self.link,
            )

        return result



from __future__ import annotations

from dataclasses import KW_ONLY, MISSING, field
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .component import Component
from .fundamental_component import FundamentalComponent

__all__ = [
    "CustomListItem",
    "HeadingListItem",
    "SimpleListItem",
]


class HeadingListItem(FundamentalComponent):
    """
    # HeadingListItem

    A simple list item with only a header.

    `HeadingListItem`s are the easiest way to create list items, which can
    take care of the most primitive task: Display a text. If your want a more
    generic list item with additional children, you can use the `SimpleListItem`.
    Or if you want to build a more complex list item, you can use the `CustomListItem`.

    Note that like with other `ListItem`s the `key` is not optional. This is to
    avoid unintentional reconciliation with other items when the list is
    updated.


    ## Attributes:

    `text`: The text to display.

    `key`: A unique key to identify the list item. This is used to avoid
        unintentional reconciliation with other items when the list is
        updated.


    ## Example:

    This minimal example will simply display a list item with the text "Click
    me!":

    ```python
    rio.HeadingListItem("Click me!", key="item1")
    ```

    `ListView`s are commonly used to display lists of dynamic length. You can easily
    achieve this by adding the children to a list first, and then unpacking that list:

    ```python
    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.HeadingListItem(
                        text=product,
                        key=product,
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    text: str
    key: str

    def __init__(
        self,
        text: str,
        *,
        key: str,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            key=key,
        )

        self.text = text


HeadingListItem._unique_id = "HeadingListItem-builtin"


class SimpleListItem(Component):
    """
    # SimpleListItem

    A simple list item with a header and optional secondary text and children.

    `SimpleListItem`s are a convenient way to create list items, which can
    take care of the most common tasks: Display a text, optional secondary text
    and even additional children (e.g. icons or buttons) to the left and right.
    Most children are optional so you can only add whichever parts you need.

    Note that like with other `ListItem`s the `key` is not optional. This is to
    avoid unintentional reconciliation with other items when the list is
    updated.


    ## Attributes

    `text`: The text to display.

    `secondary_text`: Additional text to display below the primary text. This
        text may span multiple lines (use `"\\n"` to add a line break).

    `left_child`: A component to display on the left side of the list item.

    `right_child`: A component to display on the right side of the list item.

    `on_press`: Triggered when the list item is pressed.

    `key`: A unique key to identify the list item. This is used to avoid
        unintentional reconciliation with other items when the list is
        updated.


    ## Example

    This minimal example will simply display a list item with the text "Click
    me!":

    ```python
    rio.SimpleListItem("Click me!", key="item1")
    ```

    `ListView`s are commonly used to display lists of dynamic length. You can easily
    achieve this by adding the children to a list first, and then unpacking that list:

    ```python
    import functools


    class MyComponent(rio.Component):
        products: list[str] = ["Product 1", "Product 2", "Product 3"]

        def on_press_simple_list_item(self, product: str) -> None:
            print(f"Selected {product}")

        def build(self) -> rio.Component:
            # Store all children in an intermediate list
            list_items = []

            for product in self.products:
                list_items.append(
                    rio.SimpleListItem(
                        text=product,
                        key=product,
                        left_child=rio.Icon("castle"),
                        # Note the use of `functools.partial` to pass the product
                        # to the event handler.
                        on_press=functools.partial(
                            self.on_press_simple_list_item,
                            product=product,
                        ),
                    )
                )

            # Then unpack the list to pass the children to the `ListView`
            return rio.ListView(*list_items)
    ```
    """

    text: str
    secondary_text: str
    left_child: rio.Component | None
    right_child: rio.Component | None
    on_press: rio.EventHandler[[]]
    key: str

    def __init__(
        self,
        text: str,
        *,
        key: str,
        secondary_text: str = "",
        left_child: rio.Component | None = None,
        right_child: rio.Component | None = None,
        on_press: rio.EventHandler[[]] = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            key=key,
        )

        self.text = text
        self.secondary_text = secondary_text
        self.left_child = left_child
        self.right_child = right_child
        self.on_press = on_press

    def build(self) -> rio.Component:
        children = []

        # Left child
        if self.left_child is not None:
            children.append(self.left_child)

        # Main content (text)
        text_children = [
            rio.Text(
                self.text,
                align_x=0,
            )
        ]

        if self.secondary_text:
            text_children.append(
                rio.Text(
                    self.secondary_text,
                    multiline=True,
                    style="dim",
                    align_x=0,
                )
            )

        children.append(
            rio.Column(
                *text_children,
                spacing=0.5,
                width="grow",
                align_y=0.5,  # Prevent Undefined space
            )
        )

        # Right child
        if self.right_child is not None:
            children.append(self.right_child)

        # Combine everything
        return CustomListItem(
            content=rio.Row(
                *children,
                spacing=1,
                width="grow",
            ),
            on_press=self.on_press,
            key="",
        )


class CustomListItem(FundamentalComponent):
    """
    # CustomListItem

    A list item with custom content.

    Most of the time the `SimpleListItem` will do the job. With `CustomListItems` you
    can build more complex list items. You can add any component to the list item.
    This can be e.g. a `Row`, `Column`, `Text`, `Icon`, `Image` or any other component.

    Note that like with other `ListItem`s the `key` is not optional. This is to
    avoid unintentional reconciliation with other items when the list is
    updated.


    ## Attributes:

    `content`: The content to display.

    `on_press`: Triggered when the list item is pressed.

    `key`: A unique key to identify the list item. This is used to avoid
        unintentional reconciliation with other items when the list is
        updated.


    ## Example:
    TODO: add example

    ```python

    ```
    """

    content: rio.Component
    on_press: rio.EventHandler[[]]
    key: str

    def __init__(
        self,
        content: rio.Component,
        *,
        key: str,
        on_press: rio.EventHandler[[]] = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            key=key,
        )

        self.content = content
        self.on_press = on_press

    def _custom_serialize(self) -> JsonDoc:
        return {
            "pressable": self.on_press is not None,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        if self.on_press is None:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


CustomListItem._unique_id = "CustomListItem-builtin"



"""
Implements HTTP byte serving, as per RFC 7233.

Based on a comment here:
https://github.com/tiangolo/fastapi/issues/1240#issuecomment-1055396884
"""

from pathlib import Path
from typing import *  # type: ignore

import fastapi
from fastapi import HTTPException

__all__ = [
    "range_requests_response",
]


def send_bytes_range_requests(
    file_obj: BinaryIO,
    start: int,
    end: int,
    chunk_size_in_bytes: int = 16 * 1024 * 1024,
) -> Iterator[bytes]:
    """
    Send a file in chunks using Range Requests specification RFC7233. `start`
    and `end` are inclusive as per the spec.
    """
    with file_obj as f:
        f.seek(start)
        while (pos := f.tell()) <= end:
            read_size = min(chunk_size_in_bytes, end + 1 - pos)
            yield f.read(read_size)


def parse_range_header(range_header: str, file_size: int) -> tuple[int, int]:
    try:
        h = range_header.replace("bytes=", "").split("-")
        start = int(h[0]) if h[0] != "" else 0
        end = int(h[1]) if h[1] != "" else file_size - 1
    except ValueError:
        raise HTTPException(
            fastapi.status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
            detail=f"Invalid request range header (Range: {range_header!r})",
        )

    if start > end or start < 0 or end > file_size - 1:
        raise HTTPException(
            fastapi.status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
            detail=f"Requested range is out of bounds: (Range: [{start},{end}]) (File size: {file_size}B)",
        )

    return start, end


def range_requests_response(
    request: fastapi.Request,
    file_path: Path,
    *,
    media_type: str | None = None,
) -> fastapi.responses.Response:
    """
    Returns a fastapi response which serves the given file, supporting Range
    Requests as per RFC7233 ("HTTP byte serving").

    Returns a 404 if the file does not exist.
    """

    # Get the file size. This also verifies the file exists.
    try:
        file_size_in_bytes = file_path.stat().st_size
    except FileNotFoundError:
        return fastapi.responses.Response(status_code=404)

    # Prepare response headers
    headers = {
        "accept-ranges": "bytes",
        "content-encoding": "identity",
        "content-length": str(file_size_in_bytes),
        "access-control-expose-headers": (
            "content-type, accept-ranges, content-length, content-range, content-encoding"
        ),
    }

    if media_type is not None:
        headers["content-type"] = media_type

    # Was a specific range requested?
    range_header = request.headers.get("range")
    if range_header is None:
        start = 0
        end = file_size_in_bytes - 1
        status_code = fastapi.status.HTTP_200_OK
    else:
        start, end = parse_range_header(range_header, file_size_in_bytes)
        size = end - start + 1
        headers["content-length"] = str(size)
        headers["content-range"] = f"bytes {start}-{end}/{file_size_in_bytes}"
        status_code = fastapi.status.HTTP_206_PARTIAL_CONTENT

    # Construct the response
    return fastapi.responses.StreamingResponse(
        send_bytes_range_requests(file_path.open("rb"), start, end),
        headers=headers,
        status_code=status_code,
    )



from typing import *  # type: ignore

import rio


class SkillBars(rio.Component):
    skills: dict[str, int]

    def build(self) -> rio.Component:
        rows = []

        for name, level in self.skills.items():
            rows.append(
                rio.Column(
                    rio.Text(
                        name,
                        align_x=0,
                    ),
                    rio.ProgressBar(
                        level / 10,
                    ),
                )
            )

        return rio.Card(
            rio.Column(
                *rows,
                spacing=0.5,
            ),
        )



import logging
from pathlib import Path
from typing import Any

import tomlkit
import tomlkit.exceptions


class TomlConfig:
    """
    Easy way to load and save configuration files in TOML format.

    The configuration file is loaded once, and then cached. Any modified values
    are remembered and written back in one go later on. Writing changes to the
    file is done atomically and preserves comments and formatting in the file.
    """

    def __init__(
        self,
        config_path: Path,
        *,
        default: dict[str, Any] = {},
    ):
        # Where to load the config from and save it back to
        self.config_path = config_path

        # Default values to use if the config file or any entries don't exist
        self.default = default

        # The configuration file's contents
        self._config: dict[str, Any] = {}
        self._load()

        # Keeps track of any keys which were changed and need to be written back
        self._dirty_keys: dict[tuple[str, ...], Any] = {}

    def _load(self) -> None:
        """
        Load the contents from `self.config_path` and store them in
        `self._config`.
        """
        assert not self._config

        # Try to load the toml file
        try:
            self._config = tomlkit.load(
                self.config_path.open()
            ).unwrap()  # TODO: Does this have to be done recursively?

        # No such file. Use the default values
        except FileNotFoundError:
            self._config = self.default
            return

    def save(self) -> None:
        """
        Write any changes back to the `self.config_path` file.
        """
        # Are there even any changes to write?
        if not self._dirty_keys:
            return

        # Make sure the parent directory exists
        self.config_path.parent.mkdir(parents=True, exist_ok=True)

        # Fetch an up-to-date copy of the file contents, with all formatting
        # intact
        try:
            with self.config_path.open() as f:
                new_toml_dict = tomlkit.load(f)

        # If it can't be read, preserve all known values
        except (OSError, tomlkit.exceptions.TOMLKitError) as e:
            new_toml_dict = tomlkit.TOMLDocument()

            for key, value in self._config.items():
                new_toml_dict[key] = value

        # Otherwise add just the dirty keys
        else:
            for key, value in self._dirty_keys.items():
                for subkey in key[:-1]:
                    new_toml_dict.setdefault(subkey, {})

                new_toml_dict[key[-1]] = value

        # Dump the new TOML
        with self.config_path.open("w") as f:
            tomlkit.dump(new_toml_dict, f)

        # All values have been saved and are now clean
        self._dirty_keys.clear()

    def __enter__(self) -> "TomlConfig":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # Make sure to write any changes back to the `rio.toml` file
        self.save()

    def __getitem__(self, *key: str) -> Any:
        current_dict = self._config

        for subkey in key:
            try:
                current_dict = current_dict[subkey]
            except KeyError:
                raise KeyError(key) from None

        return current_dict

    def __setitem___(self, *key: str, value: Any) -> None:
        """
        Sets the value of a key in configuration. The value is not written back
        to disk until `write()` is called.
        """
        # Add the value to the config
        current_dict = self._config

        for ii, subkey in enumerate(key[:-1]):
            current_dict = current_dict.setdefault(subkey, {})

            if not isinstance(current_dict, dict):
                key_until_now = key[: ii + 1]
                raise ValueError(f"`{key_until_now}` is not a dictionary")

        current_dict[key[-1]] = value

        # Remember that this key was changed
        self._dirty_keys[key] = value



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import Any

from uniserde import JsonDoc

import rio

from .fundamental_component import KeyboardFocusableFundamentalComponent

__all__ = [
    "MultiLineTextInput",
    "MultiLineTextInputChangeEvent",
    "MultiLineTextInputConfirmEvent",
]


@dataclass
class MultiLineTextInputChangeEvent:
    text: str


@dataclass
class MultiLineTextInputConfirmEvent:
    text: str


class MultiLineTextInput(KeyboardFocusableFundamentalComponent):
    """
    # MultiLineTextInput

    A user-editable text field.

    It's similar to `TextInput`, but it allows the user to enter multiple lines of text.

    `MultiLineTextInput` allows the user to enter a short text. The text can either be
    shown in plain text or other sensitive information.


    ## Attributes:

    `text`: The text currently entered by the user.

    `label`: A short text to display next to the text input.

    `is_sensitive`: Whether the text input should respond to user input.

    `is_valid`: Visually displays to the user whether the current text is
            valid. You can use this to signal to the user that their input needs
            to be changed.

    `on_change`: Triggered when the user changes the text.

    `on_confirm`: Triggered when the user explicitly confirms their input,
            such as by pressing the "Enter" key. You can use this to trigger
            followup actions, such as logging in or submitting a form.


    ## Example:

    A minimal example of `MultiLineTextInput` with a default text of "" will be shown:
    `Note`: The text will not be updated if the user changes the text in the input field.

    ```python
    rio.MultiLineTextInput(text="")
    ```

    You can easily `bind` state variables to track changes. If you want to make your
    `MultiLineTextInput` more responsive, you can easily achieve this by adding a
    lambda function call to e.g. `on_change`:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def build(self) -> rio.Component:
            return rio.MultiLineTextInput(
                text=self.bind().text,  # state binding
                label="Write your comments here",
                on_change=lambda event: print(event.text),
            )
    ```

    If you are not using state bindings, you can also use a lambda function for updating
    the input text:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def build(self) -> rio.Component:
            return rio.MultiLineTextInput(
                text=self.text,
                label="Write your comments here",
                on_change=lambda event: setattr(
                    self, "text", event.text
                ),  # Update the text
            )
    ```

    You can also use a method for updating the input text and do whatever you want.
    `Note` that methods are handy if you want to do more than just updating the input
    text. For example run async code or update other components based on the input text:

    ```python
    class MyComponent(rio.Component):
        text: str = ""

        def on_change_update_text(self, event: rio.MultiLineTextInputChangeEvent):
            self.text = event.text
            # You can do whatever you want in here

        def build(self) -> rio.Component:
            return rio.MultiLineTextInput(
                text=self.text,
                label="Write your comments here",
                on_change=self.on_change_update_text,
            )
    ```
    """

    text: str = ""
    _: KW_ONLY
    label: str = ""
    is_sensitive: bool = True
    is_valid: bool = True
    on_change: rio.EventHandler[MultiLineTextInputChangeEvent] = None
    on_confirm: rio.EventHandler[MultiLineTextInputConfirmEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"text"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "text" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `MultiLineTextInput.text` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["text"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, str), new_value
            await self.call_event_handler(
                self.on_change,
                MultiLineTextInputChangeEvent(new_value),
            )

        self._apply_delta_state_from_frontend(delta_state)

    async def _on_message(self, msg: Any) -> None:
        # Listen for messages indicating the user has confirmed their input
        #
        # In addition to notifying the backend, these also include the input's
        # current value. This ensures any event handlers actually use the up-to
        # date value.
        assert isinstance(msg, dict), msg

        self._apply_delta_state_from_frontend({"text": msg["text"]})

        await self.call_event_handler(
            self.on_change,
            MultiLineTextInputChangeEvent(self.text),
        )

        await self.call_event_handler(
            self.on_confirm,
            MultiLineTextInputConfirmEvent(self.text),
        )

        # Refresh the session
        await self.session._refresh()


MultiLineTextInput._unique_id = "MultiLineTextInput-builtin"



from __future__ import annotations

from collections.abc import Callable
from dataclasses import KW_ONLY, field
from typing import *  # type: ignore

import rio

from .component import Component

__all__ = [
    "PageView",
]


def default_fallback_build(sess: rio.Session) -> rio.Component:
    thm = sess.theme

    return rio.Column(
        rio.Row(
            rio.Icon(
                "material/error",
                fill="warning",
                width=4,
                height=4,
            ),
            rio.Text(
                "This page does not exist",
                style=rio.TextStyle(
                    font_size=3,
                    fill=thm.warning_palette.background,
                ),
            ),
            spacing=2,
            align_x=0.5,
        ),
        rio.Text(
            "The URL you have entered does not exist on this website. Double check your spelling, or try navigating to the home page.",
            multiline=True,
        ),
        rio.Button(
            "Take me home",
            on_press=lambda: sess.navigate_to("/"),
        ),
        spacing=3,
        width=20,
        align_x=0.5,
        align_y=0.35,
    )


class PageView(Component):
    """
    # PageView

    Placeholders for pages.

    Rio apps can consist of many pages. You might have a welcome page, a
    settings page, a login, and so on. Page views act as placeholders, that
    don't have an appearance of their own, but instead look up the currently
    active page, and display that.


    ## Attributes:

    `fallback_build`: A callback that is called if the current URL does not
        correspond to any page in the application. It should return a Rio
        component that is displayed instead. If not specified Rio will
        display a default error page.


    ## Example:

    A minimal example:

    ```python
    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my page!"),
            rio.PageView(
                width="grow",
                height="grow",
            ),
        ),
        pages=[
            rio.Page(
                "",
                build=lambda: rio.Text("This is the home page"),
            ),
            rio.Page(
                "subpage",
                build=lambda: rio.Text("This is a subpage"),
            ),
        ],
    )

    app.run_in_browser()
    ```

    This will display "This is the home page" when navigating to the root URL,
    but "This is a subpage" when navigating to "/subpage". Note that on both
    pages the text "Welcome to my page!" is displayed above the page content.
    That's because it's not part of the `PageView`.

    # TODO: Link to the routing/multipage how-to page

    """

    _: KW_ONLY

    fallback_build: Callable[[], rio.Component] | None = None

    # How many other PageViews are above this one in the component tree. Zero
    # for top-level PageViews, 1 for the next level, and so on.
    #
    # Initialized in `__post_init__`.
    level: int = field(init=False)

    def __post_init__(self) -> None:
        self.session._page_views.add(self)

        # Determine how many `PageView`s are above this one. This can never
        # change, because `PageView`s cannot possibly be moved into or out of
        # other `PageView`s by the reconciler, since they don't accept children.
        self.level = self._find_page_view_level()

    def _find_page_view_level(self) -> int:
        """
        Follow the chain of `_weak_creator_` references to find how deep in the
        hierarchy of PageViews this one is. Returns 0 for a top-level PageView,
        1 for a PageView inside one other PageView, etc.

        Normally, following `_weak_creator_` references is not safe, because it
        can skip over other components, and even across build boundaries.
        However, since `PageView`s don't accept children, it isn't possible for
        them to ever be moved into or out of other `PageView`s by the
        reconciler.
        """
        cur_parent = self._weak_creator_()

        while True:
            # No more parents - this is the root
            if cur_parent is None:
                return 0

            # Found it
            if isinstance(cur_parent, PageView):
                return cur_parent.level + 1

            # Chain up
            cur_parent = cur_parent._weak_creator_()

    def build(self) -> rio.Component:
        # Build the active page
        #
        # Because the build function is being called inside of Rio, the
        # component is mistaken of being internal to Rio. Take care to fix that.
        try:
            page = self.session._active_page_instances[self.level]
        except IndexError:
            if self.fallback_build is None:
                result = default_fallback_build(self.session)
            else:
                result = self.fallback_build()
                result._rio_internal_ = False
        else:
            result = page.build()
            result._rio_internal_ = False

        return result
